import logging
import multiprocessing
from collections import deque
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import polars as pl
from numba import njit
from scipy.sparse import csr_matrix, issparse
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.spatial import distance_matrix
from skimage.measure import approximate_polygon
from skimage.morphology import dilation, disk, erosion
from tqdm import tqdm

from ..g4x_viewer import CellMasksSchema_pb2 as CellMasksSchema
from .edit_bin import DEFAULT_COLOR, hex2rgb, update_colormap
from .segment import get_cell_ids
from .workflow import OutSchema, workflow

if TYPE_CHECKING:
    from ..models import G4Xoutput


@workflow
def init_bin_file(
    g4x_obj: 'G4Xoutput',
    out_dir: str | Path,
    seg_mask: np.ndarray | None = None,
    single_cell_dir: str | Path | None = None,
    *,
    n_threads: int = 4,
    logger: logging.Logger | None = None,
) -> None:
    out_tree = OutSchema(out_dir, subdirs=['g4x_viewer'])
    out_file = out_tree.g4x_viewer / f'{g4x_obj.sample_id}_segmentation.bin'

    adata = g4x_obj.load_adata(
        single_cell_dir=single_cell_dir or g4x_obj.data_dir / 'single_cell_data',
        load_clustering=False,
        remove_nontargeting=False,
    )

    if seg_mask is None:
        logger.info('Loading default segmentation mask.')
        seg_mask = g4x_obj.load_segmentation()

    logger.info('Loading clustering information.')
    obs_df = prepare_metadata(sample_id=g4x_obj.sample_id, adata=adata, seg_mask=seg_mask)
    print(f'Number of cells to process: {obs_df.shape[0]}')

    logger.info('Making polygons.')
    cell_ids, pq_args = initialize_segmentation_data(obs_df, seg_mask)

    logger.debug('Adding single-cell metadata.')
    obs_df, gene_names, gex = add_singlecell_info(obs_df, adata, cell_ids)

    ## do conversion
    cell_masks = CellMasksSchema.CellMasks()
    cell_masks.metadata.geneNames.extend(gene_names)

    logger.info('Setting default color.')
    cluster_palette = {}
    cluster_palette['unassigned'] = hex2rgb(DEFAULT_COLOR)

    update_colormap(cell_masks, cluster_palette)

    protein_values = None
    if g4x_obj.includes_protein:
        protein_names = g4x_obj.proteins
        protein_list = [prot + '_intensity_mean' for prot in protein_names]

        cell_masks.metadata.proteinNames.extend(protein_names)

        obs_df = obs_df.with_columns([pl.col(prot).fill_null(0).cast(pl.Int64) for prot in protein_list])
        protein_values = obs_df.select(protein_list).to_numpy()

    logger.info('Refining polygons.')
    with multiprocessing.Pool(processes=n_threads) as pool:
        polygons = pool.starmap(refine_polygon, pq_args)

    logger.info('Adding individual cells.')
    for i, row in enumerate(tqdm(obs_df.iter_rows(named=True), total=obs_df.height, desc='Adding individual cells.')):
        output_mask_data = cell_masks.cellMasks.add()

        cell_polygon_pts = [sub_coord for coord in polygons[i] for sub_coord in coord[::-1]]
        output_mask_data.vertices.extend(cell_polygon_pts + cell_polygon_pts[:2])

        output_mask_data.cellId = str(row['cell_id'])
        output_mask_data.area = int(row['area'])
        output_mask_data.totalCounts = int(row['total_counts'])
        output_mask_data.totalGenes = int(row['n_genes_by_counts'])
        output_mask_data.clusterId = str(row['cluster_id'])
        output_mask_data.umapValues.umapX = float('nan')
        output_mask_data.umapValues.umapY = float('nan')

        start = gex.indptr[i]
        end = gex.indptr[i + 1]
        indices = gex.indices[start:end]
        values = gex.data[start:end]
        output_mask_data.nonzeroGeneIndices.extend(indices.tolist())
        output_mask_data.nonzeroGeneValues.extend(values.astype(int).tolist())

        if protein_values is not None:
            output_mask_data.proteinValues.extend(protein_values[i].astype(int).tolist())

    ## write to file
    with open(out_file, 'wb') as file:
        file.write(cell_masks.SerializeToString())

    logger.debug(f'G4X-viewer bin --> {out_file}')


def prepare_metadata(sample_id: str, adata, seg_mask: np.ndarray) -> pl.DataFrame:
    obs_df = get_cell_ids(sample_id=sample_id, mask=seg_mask)

    metadata_df = pl.from_pandas(adata.obs)
    obs_df = obs_df.join(metadata_df, left_on='label', right_on='cell_id', how='left').rename({'label': 'cell_id'})

    obs_df = obs_df.with_columns(pl.lit('unassigned').alias('cluster_id'))

    return obs_df


def initialize_segmentation_data(obs_df: pl.DataFrame, seg_mask: np.ndarray) -> tuple[list[str], list[tuple]]:
    ## we create polygons to define the boundaries of each cell mask
    border = get_border(seg_mask)
    seg_mask[border > 0] = 0
    eroded_mask = erosion(seg_mask, disk(1))
    outlines = seg_mask - eroded_mask
    sparse_matrix = csr_matrix(outlines)
    del seg_mask, border, eroded_mask, outlines

    nonzero_values = sparse_matrix.data
    nonzero_row_indices, nonzero_col_indices = sparse_matrix.nonzero()
    sorted_indices = np.argsort(nonzero_values)
    sorted_nonzero_values = nonzero_values[sorted_indices]
    sorted_rows = nonzero_row_indices[sorted_indices]
    sorted_cols = nonzero_col_indices[sorted_indices]

    cell_ids = obs_df['cell_id'].to_list()
    segmentation_labels = obs_df['segmentation_label'].to_list()

    centroid_y = obs_df['cell_x'].to_list()
    centroid_x = obs_df['cell_y'].to_list()

    pq_args = [
        (seg_label, sorted_nonzero_values, sorted_rows, sorted_cols)
        for seg_label, _, _ in zip(segmentation_labels, centroid_x, centroid_y)
    ]

    return cell_ids, pq_args


def add_singlecell_info(obs_df: pl.DataFrame, adata, cell_ids: list[str]):
    # filter adata
    adata = adata[cell_ids]
    gene_names = adata.var_names

    gex = adata.X
    if not issparse(gex):
        print('Converting dense expression matrix to CSR format.')
        gex = csr_matrix(gex)
    else:
        gex = gex.tocsr()

    del adata

    obs_df = obs_df.cast({'total_counts': pl.UInt32, 'n_genes_by_counts': pl.Int32})

    # add_singlecell_area
    area_cols = [c for c in obs_df.columns if 'area' in c]
    if not len(area_cols):
        raise ValueError('No area column found')

    if len(area_cols) == 1:
        area_select = area_cols[0]
    else:
        expanded_area_cols = [f for f in area_cols if 'expanded' in f]
        if len(expanded_area_cols) != 1:
            raise ValueError(
                f'No expanded area column found among multiple area columns: {area_cols}.'
                if not expanded_area_cols
                else f'Multiple expanded area columns found: {expanded_area_cols}'
            )
        print(f'Multiple area columns found. Using {expanded_area_cols[0]}.')
        area_select = expanded_area_cols[0]

    print(f'Using area column: {area_select}')

    obs_df = obs_df.cast({area_select: pl.Int32}).rename({area_select: 'area'})

    return obs_df, gene_names, gex


def refine_polygon(
    cell_id: int, sorted_nonzero_values_ref: np.ndarray, sorted_rows_ref: np.ndarray, sorted_cols_ref: np.ndarray
) -> np.ndarray:
    start_idx, end_idx = get_start_stop_idx(sorted_nonzero_values_ref, cell_id)
    points = np.vstack((sorted_rows_ref[start_idx:end_idx], sorted_cols_ref[start_idx:end_idx])).T
    return pointsToSingleSmoothPath(points, tolerance=2.0)


@njit
def get_start_stop_idx(arr: np.ndarray, k: int) -> tuple[int, int]:
    start_idx = np.searchsorted(arr, k, side='left')
    end_idx = np.searchsorted(arr, k, side='right')
    return start_idx, end_idx


def pointsToSingleSmoothPath(points: np.ndarray, tolerance: float) -> np.ndarray:
    # Calculate the distance matrix
    dist_matrix = distance_matrix(points, points)

    # Create a sparse matrix for the MST calculation
    sparse_matrix = csr_matrix(dist_matrix)

    # Compute the MST
    mst = minimum_spanning_tree(sparse_matrix).toarray()

    adj_list, adj_list_pos = createAdjacencyList_numba(mst)
    adj_list = {row: list(adj_list[row, :pos]) for row, pos in enumerate(adj_list_pos) if pos}

    longest_path = computeLongestPath(adj_list)
    bestPath = indicesToArray(points, longest_path)

    simplified_path = simplify_polygon(bestPath, tolerance=tolerance)

    return simplified_path


def computeLongestPath(adj_list: dict[int, list[int]]) -> list[int]:
    endpoints = returnEndpoints(adj_list)
    longest_path = []
    max_length = 0
    # Use a dictionary to cache paths and avoid recomputation
    path_cache = {}
    # Compute distances between all pairs of endpoints
    for i in range(len(endpoints)):
        for j in range(i + 1, len(endpoints)):
            if (endpoints[i], endpoints[j]) not in path_cache:
                path = bfs_path(endpoints[i], endpoints[j], adj_list)
                path_cache[(endpoints[i], endpoints[j])] = path
            else:
                path = path_cache[(endpoints[i], endpoints[j])]
            if len(path) > max_length:
                max_length = len(path)
                longest_path = path

    return longest_path


def returnEndpoints(adj_list: dict[int, list[int]], adjacency: int = 2) -> list[int]:
    # Identify endpoints of the MST
    endpoints = [node for node in adj_list if len(adj_list[node]) == adjacency]

    return endpoints


def bfs_path(start: int, end: int, adj_list: dict[int, list[int]]) -> list[int]:
    queue = deque([(start, [start])])
    visited = set()
    while queue:
        current, path = queue.popleft()
        if current == end:
            return path
        if current in visited:
            continue
        visited.add(current)
        for neighbor in adj_list[current]:
            if neighbor not in visited:
                queue.append((neighbor, path + [neighbor]))
    return []


@njit
def createAdjacencyList_numba(mst: np.ndarray) -> tuple:
    """
    Create an adjacency list from a minimum spanning tree (MST) using Numba for performance optimization.

    Parameters:
    mst (numpy.ndarray): The minimum spanning tree represented as a 2D numpy array.

    Returns:
    tuple: A tuple containing:
        - adj_list (numpy.ndarray): An array where each row contains the adjacent nodes for each node.
        - adj_list_pos (numpy.ndarray): An array containing the number of adjacent nodes for each node.
    """
    n = mst.shape[0]
    adj_list = np.zeros((n, n * 2), dtype=np.uint32)
    adj_list_pos = np.zeros(n, dtype=np.uint32)
    for i in range(n):
        for j in range(n):
            if mst[i, j] != 0 or mst[j, i] != 0:
                adj_list[i, adj_list_pos[i]] = j
                adj_list_pos[i] += 1
                adj_list[j, adj_list_pos[j]] = i
                adj_list_pos[j] += 1

    return adj_list, adj_list_pos


def indicesToArray(points: np.ndarray, longest_path: list[int]) -> np.ndarray:
    pth = []

    for j in range(len(longest_path)):
        pth.append([points[longest_path[j], 0], points[longest_path[j], 1]])

    return np.array(pth)


def simplify_polygon(points: np.ndarray, tolerance: float) -> np.ndarray:
    """
    Simplify a series of points representing a polygon using scikit-image's
    approximate_polygon. The tolerance controls how aggressively the polygon
    is simplified (in pixel units).
    """
    if len(points) <= 2:
        return points

    # If the first and last points are not the same, append the first to the end
    # to ensure the polygon is "closed" for approximation (optional).
    if not np.array_equal(points[0], points[-1]):
        points = np.vstack([points, points[0]])

    # Perform polygon simplification
    simplified = approximate_polygon(points, tolerance=tolerance)

    # If approximate_polygon returns only the closed ring, remove the last point
    # to avoid duplication in your pipeline. (approx_polygon returns a closed ring
    # by repeating the first point at the end.)
    if len(simplified) > 2 and np.array_equal(simplified[0], simplified[-1]):
        simplified = simplified[:-1]

    return simplified


def get_border(mask: np.ndarray, s: int = 1) -> np.ndarray:
    d = dilation(mask, disk(s))
    border = (mask != d).astype(np.uint8)
    return border



import json
import os
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Literal

import glymur
import numpy as np
import pandas as pd
import polars as pl
import tifffile
from anndata import AnnData, read_h5ad
from matplotlib.pyplot import imread

from . import utils
from .schemas import validate


@dataclass()
class G4Xoutput:
    """
    Container for managing and processing data from a G4X run.

    This class initializes and loads metadata, image dimensions, transcript and protein panels for downstream analysis of G4X output data.
    It provides methods to load images, segmentations, transcript data, and interact with single-cell and spatial analysis pipelines.

    Parameters
    ----------
    data_dir : str
        Path pointing to a G4X sample output directory. This should contain all run-related files including metadata,
        segmentation masks, panels, and feature matrices.
    sample_id : str, optional
        The sample ID to associate with the run. If not provided, it will be inferred from the `data_dir` path.

    Attributes
    ----------
    run_meta : dict
        Metadata dictionary loaded from `run_meta.json`.
    shape : tuple
        Image shape (height, width) as inferred from the segmentation mask.
    transcript_panel_dict : dict
        Mapping of transcript genes to panel types, if the transcript panel is present.
    protein_panel_dict : dict
        Mapping of proteins to panel types, if the protein panel is present.
    genes : list of str
        List of transcript gene names (only if transcript panel exists).
    proteins : list of str
        List of protein names (only if protein panel exists).

    Notes
    -----
    On instantiation, this class performs the following:
      - Loads metadata from `run_meta.json`.
      - Loads the shape of the segmentation mask.
      - Parses transcript and protein panel files (if present).
    """

    data_dir: str
    sample_id: str | None = None

    def __post_init__(self):
        self.data_dir = Path(self.data_dir)

        if self.sample_id is None:
            self.sample_id = self.infer_sample_id()

        with open(self.data_dir / 'run_meta.json', 'r') as f:
            self.run_meta = json.load(f)

        self.set_meta_attrs()

        if self.transcript_panel:
            try:
                transcript_panel = (
                    utils.parse_input_manifest(self.data_dir / 'transcript_panel.csv')
                    .unique('gene_name')
                    .sort('gene_name')
                )
                self.transcript_panel_dict = dict(zip(transcript_panel['gene_name'], transcript_panel['panel_type']))
                self.genes = transcript_panel['gene_name'].to_list()
            except Exception:
                self.transcript_panel_dict = {}
                self.genes = []

        if self.protein_panel:
            protein_panel = pl.read_csv(self.data_dir / 'protein_panel.csv').sort('target')
            self.protein_panel_dict = dict(zip(protein_panel['target'], protein_panel['panel_type']))
            self.proteins = protein_panel['target'].to_list()

    # region dunder
    def __repr__(self):
        machine_num = self.machine.removeprefix('g4-').lstrip('0')
        mac_run_id = f'G{machine_num.zfill(2)}-{self.run_id}'
        gap = 16
        repr_string = f'G4X-data @ {self.data_dir}\n'

        shp = (np.array(self.shape) * 0.3125) / 1000

        repr_string += f'{"Sample ID":<{gap}} - {self.sample_id} of {mac_run_id}, {self.fc}\n'
        repr_string += f'{"imaged area":<{gap}} - ({shp[1]:.2f} x {shp[0]:.2f}) mm\n'
        repr_string += f'{"software version":<{gap}} - {self.software_version}\n\n'

        panels = [
            ('Transcript panel', len(self.genes), 'genes', self.genes)
            if self.includes_transcript
            else (None, 0, '', []),
            ('Protein panel', len(self.proteins), 'proteins', self.proteins)
            if self.includes_protein
            else (None, 0, '', []),
        ]

        # Step 1: compute lengths of "<count> <label>"
        pre_bracket_lengths = [
            len(str(count)) + 2 + len(label)  # e.g., "128 genes"
            for (_, count, label, _) in panels
        ]

        # Step 2: max width to align the `[`
        max_pre = max(pre_bracket_lengths)

        def format_panel(title, count, label, items):
            return f'{title:<{gap}} - {count} {label:<{max_pre - len(str(count)) - 1}}[{", ".join(items[0:5])} ... ]\n'

        if self.includes_transcript:
            repr_string += format_panel(*panels[0])

        if self.includes_protein:
            repr_string += format_panel(*panels[1])

        return repr_string

    # region properties
    @property
    def includes_protein(self) -> bool:
        return self.protein_panel != []

    @property
    def includes_transcript(self) -> bool:
        return self.transcript_panel != []

    @property
    def transcript_table_path(self) -> Path:
        return self.data_dir / 'rna' / 'transcript_table.csv.gz'

    @property
    def feature_table_path(self) -> Path:
        return self.data_dir / 'rna' / 'raw_features.parquet'

    @property
    def segmentation_path(self) -> Path:
        return self.data_dir / 'segmentation' / 'segmentation_mask.npz'

    @property
    def feature_mtx_path(self) -> Path:
        return self.data_dir / 'single_cell_data' / 'feature_matrix.h5'

    @property
    def bead_mask_path(self) -> Path:
        return self.data_dir / 'protein' / 'bead_mask.npz'

    @property
    def use_cyto(self):
        h_and_e_images = []
        for img in (self.data_dir / 'h_and_e').glob('*.jp2'):
            h_and_e_images.append(img.stem)

        return 'cytoplasmic' in h_and_e_images

    def set_meta_attrs(self):
        static_attrs = [
            'platform',
            'machine',
            'run_id',
            'fc',
            'lane',
            'software',
            'software_version',
            'transcript_panel',
            'protein_panel',
        ]

        for k in static_attrs:
            setattr(self, k, self.run_meta.get(k, None))

        self.shape = glymur.Jp2k(self.data_dir / 'h_and_e' / 'nuclear.jp2').shape

    def infer_sample_id(self) -> str:
        summs = list(self.data_dir.glob('summary_*.html'))

        failure = 'Could not determine sample_id:'
        if len(summs) == 0:
            raise validate.ValidationError(f'{failure} "summary_{{sample_id}}.html" missing')
        elif len(summs) > 1:
            raise validate.ValidationError(f'{failure} Multiple "summary_{{sample_id}}.html" files found')
        else:
            summary_file = summs[0]
            sample_id = summary_file.stem.removeprefix('summary_')
        return sample_id

    def validate(self, details: bool = False) -> None:
        report_style = 'short' if details else False
        _ = validate.validate_g4x_data(
            self.data_dir, schema_name='base_schema', formats={'sample_id': self.sample_id}, report=report_style
        )

    # region methods
    def load_adata(
        self, *, single_cell_dir: str | None = None, remove_nontargeting: bool = True, load_clustering: bool = True
    ) -> AnnData:
        if single_cell_dir is not None:
            single_cell_dir = self.data_dir / single_cell_dir

        adata = read_h5ad(single_cell_dir / self.feature_mtx_path.name)

        adata.obs_names = adata.obs['cell_id']
        adata.var_names = adata.var['gene_id']

        adata.obs['sample_id'] = adata.uns['sample_id'] = self.sample_id
        adata.uns['software_version'] = self.software_version

        if remove_nontargeting:
            adata = adata[:, adata.var.query(" probe_type == 'targeting' ").index].copy()

        if load_clustering:
            df = pd.read_csv(single_cell_dir / 'clustering_umap.csv.gz', index_col=0, header=0)
            adata.obs = adata.obs.merge(df, how='left', left_index=True, right_index=True)
            umap_key = '_'.join(sorted([x for x in adata.obs.columns if 'X_umap' in x])[0].split('_')[:-1])
            adata.obsm['X_umap'] = adata.obs[[f'{umap_key}_1', f'{umap_key}_2']].to_numpy(dtype=float)

            # convert clustering columns to categorical
            for col in adata.obs.columns:
                if 'leiden' in col:
                    adata.obs[col] = adata.obs[col].astype('category')

        adata.obs_names = f'{self.sample_id}-' + adata.obs['cell_id'].str.split('-').str[1]
        return adata

    def load_image_by_type(
        self,
        image_type: Literal['protein', 'h_and_e', 'nuclear', 'eosin'],
        *,
        thumbnail: bool = False,
        protein: str | None = None,
        cached: bool = False,
    ) -> np.ndarray:
        if image_type == 'protein':
            if not self.protein_panel:
                print('No protein results.')
                return None
            if protein is None or protein not in self.proteins:
                print(f'{protein} not in protein panel.')
                return None
            pattern = f'{protein}_thumbnail.*' if thumbnail else f'{protein}.*'
            directory = 'protein'
        else:
            pattern_base = {'h_and_e': 'h_and_e', 'nuclear': 'nuclear', 'eosin': 'eosin'}.get(image_type)

            if image_type == 'eosin' and self.use_cyto:
                pattern_base = 'cytoplasmic'

            if not pattern_base:
                print(f'Unknown image type: {image_type}')
                return None

            pattern = f'{pattern_base}_thumbnail.*' if thumbnail else f'{pattern_base}.*'
            directory = 'h_and_e'

        if cached:
            return self._load_image_cached(self.data_dir, directory, pattern)
        else:
            return self._load_image(self.data_dir, directory, pattern)

    def load_protein_image(self, protein: str, thumbnail: bool = False, cached: bool = False) -> np.ndarray:
        return self.load_image_by_type('protein', thumbnail=thumbnail, protein=protein, cached=cached)

    def load_he_image(self, thumbnail: bool = False, cached: bool = False) -> np.ndarray:
        return self.load_image_by_type('h_and_e', thumbnail=thumbnail, cached=cached)

    def load_nuclear_image(self, thumbnail: bool = False, cached: bool = False) -> np.ndarray:
        return self.load_image_by_type('nuclear', thumbnail=thumbnail, cached=cached)

    def load_eosin_image(self, thumbnail: bool = False, cached: bool = False) -> np.ndarray:
        return self.load_image_by_type('eosin', thumbnail=thumbnail, cached=cached)

    def load_segmentation(self, expanded: bool = True, key: str | None = None) -> np.ndarray:
        from .modules.segment import try_load_segmentation

        arr = np.load(self.segmentation_path)
        available_keys = list(arr.keys())

        if 'nuclei' and 'nuclei_exp' in available_keys:
            key = 'nuclei_exp' if expanded else 'nuclei'

        return try_load_segmentation(cell_labels=self.segmentation_path, expected_shape=self.shape, labels_key=key)

    def load_bead_mask(self) -> np.ndarray:
        return np.load(self.bead_mask_path)['bead_mask']

    # TODO: this is not used anywhere, consider removing. it's also broken
    def load_feature_table(
        self,
        *,
        return_polars: bool = True,
        lazy: bool = False,
        columns: list[str] | None = None,
        alt_file_path: str | None = None,
    ) -> pd.DataFrame | pl.DataFrame | pl.LazyFrame:
        file_path = self.feature_table_path if alt_file_path is None else alt_file_path
        return self._load_table(file_path, return_polars, lazy, columns)

    def load_transcript_table(
        self,
        *,
        return_polars: bool = True,
        lazy: bool = False,
        columns: list[str] | None = None,
        alt_file_path: str | None = None,
    ) -> pd.DataFrame | pl.DataFrame | pl.LazyFrame:
        file_path = self.transcript_table_path if alt_file_path is None else alt_file_path
        return self._load_table(file_path, return_polars, lazy, columns)

    def list_content(self, subdir=None):
        if subdir is None:
            subdir = ''

        list_path = self.data_dir / subdir
        output = os.listdir(list_path)

        contents = {'dirs': [], 'files': []}
        for item in output:
            if os.path.isdir(list_path / item):
                contents['dirs'].append(item)
            if os.path.isfile(list_path / item):
                contents['files'].append(item)

        return contents

    # region internal
    @staticmethod
    def _load_image_base(data_dir: str, parent_directory: str, pattern: str) -> tuple[np.ndarray, float, float]:
        img_path = next((data_dir / parent_directory).glob(pattern), None)
        if img_path is None:
            raise FileNotFoundError(f'No file matching {pattern} found.')
        if img_path.suffix == '.jp2' or img_path.suffix == '.jpg':
            img = glymur.Jp2k(img_path)[:]
        elif img_path.suffix == '.png':
            img = imread(img_path)
        else:
            img = tifffile.imread(img_path)
        return img

    @staticmethod
    @lru_cache(maxsize=None)
    def _load_image_cached(data_dir: str, parent_directory: str, pattern: str) -> tuple[np.ndarray, float, float]:
        return G4Xoutput._load_image_base(data_dir, parent_directory, pattern)

    @staticmethod
    def _load_image(data_dir: str, parent_directory: str, pattern: str) -> tuple[np.ndarray, float, float]:
        return G4Xoutput._load_image_base(data_dir, parent_directory, pattern)

    def _load_table(
        self, file_path: str, return_polars: bool = True, lazy: bool = False, columns: list[str] | None = None
    ) -> pd.DataFrame | pl.DataFrame | pl.LazyFrame:
        file_path = Path(file_path)
        if lazy:
            if file_path.suffix == '.parquet':
                reads = pl.scan_parquet(file_path)
            else:
                reads = pl.scan_csv(file_path)
        else:
            if file_path.suffix == '.parquet':
                reads = pl.read_parquet(file_path)
            else:
                reads = pl.read_csv(file_path)
        if columns:
            reads = reads.select(columns)
        if not return_polars:
            reads = reads.collect().to_pandas()
        return reads

    def _clear_image_cache(self):
        """Evict all cached images so subsequent calls re-read from disk."""
        self._load_image.cache_clear()



from __future__ import annotations

import gzip
import logging
import os
import shutil
from datetime import datetime
from pathlib import Path

import pandas as pd
import polars as pl

DEFAULT_THREADS = max(1, (os.cpu_count() // 2 or 4))
PROBE_PATTERN = r'^(.*?)-([ACGT]{2,30})-([^-]+)$'

primer_read_map = {
    'SP1': 1,
    'm7a': 2,
    'm9a': 3,
    'm6a': 4,
    'm8a': 5,
    'm3a': 6,
}


def validate_path(path_str, must_exist=True, is_dir_ok=True, is_file_ok=True, resolve_path=False) -> Path:
    path = Path(path_str)  # .expanduser().resolve()
    if resolve_path:
        path = path.resolve()

    if must_exist and not path.exists():
        raise FileNotFoundError(f'Path does not exist: {path}')

    if path.exists():
        if path.is_dir() and not is_dir_ok:
            raise ValueError(f'Expected a file but got a directory: {path}')
        if path.is_file() and not is_file_ok:
            raise ValueError(f'Expected a directory but got a file: {path}')

    if not path.exists() and not must_exist:
        if not is_file_ok:
            path.mkdir(parents=True, exist_ok=True)
        else:
            parent = path.parent
            parent.mkdir(parents=True, exist_ok=True)

    return path


def delete_existing(outfile: str | Path) -> None:
    outfile = Path(outfile)
    if outfile.exists() or outfile.is_symlink():
        outfile.unlink()


def gzip_file(outfile: str | Path, remove_original: bool = False) -> None:
    outfile = Path(outfile)

    delete_existing(outfile.with_suffix('.csv.gz'))
    with open(outfile, 'rb') as f_in:
        with gzip.open(outfile.with_suffix('.csv.gz'), 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)

    if remove_original:
        os.remove(outfile)


def write_csv_gz(df, path, keep_uncompressed: bool = False):
    if isinstance(df, pl.LazyFrame):
        df.sink_csv(path)
    if isinstance(df, pl.DataFrame):
        df.write_csv(path)
    if isinstance(df, pd.DataFrame):
        df.to_csv(path, index=False)

    gzip_file(path, remove_original=not keep_uncompressed)


def create_custom_out(
    out_dir: Path | str,
    parent_folder: Path | str | None = None,
    file_name: str | None = None,
) -> Path:
    custom_out = Path(out_dir) / parent_folder

    # Ensure the directory exists
    custom_out.mkdir(parents=True, exist_ok=True)

    # Prepare output file path
    outfile = custom_out / file_name

    delete_existing(outfile)

    return outfile


# region logger
def verbose_to_log_level(verbose: int) -> int:
    """
    returns a logging level based on verbose integer:
    0 == WARNING
    1 == REPORT
    2 == INFO
    any other integer == DEBUG
    """
    mapping = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }
    return mapping.get(verbose, logging.DEBUG)


class TruncatingFormatter(logging.Formatter):
    def __init__(self, *args, name_max=20, **kwargs):
        super().__init__(*args, **kwargs)
        self.name_max = name_max

    def format(self, record):
        if len(record.name) > self.name_max:
            end = self.name_max - len('...' + 'XX')
            record.name = record.name[:end] + '...' + record.name[-2:]
        # else:
        #     fill = '.' * (self.name_max - len(record.name))
        #     record.name = record.name + fill

        return super().format(record)


def setup_logger(
    logger_name: str,
    *,
    stream_logger: bool = True,
    stream_level: int = 2,
    file_logger: bool = True,
    file_level: int = 2,
    file_mode: str = 'a',
    out_dir: Path | str | None = None,
    format: str | None = None,
    clear_handlers: bool = True,
) -> logging.Logger:
    """
    Sets up a logger with configurable stream and file handlers.

    Parameters
    ----------
    logger_name : str
        Name of the logger.
    stream_logger : bool, optional
        Whether to enable logging to the console (default is True).
    stream_level : int, optional
        Logging level for the stream handler (default is logging.DEBUG).
    file_logger : bool, optional
        Whether to enable logging to a file (default is False).
    file_level : int, optional
        Logging level for the file handler (default is logging.DEBUG).
    file_mode : str, optional
        File mode for the log file. Common options: 'a' for append, 'w' for overwrite (default is 'a').
    out_dir : Path or str, optional
        Directory where the log file will be saved. Required if `file_logger` is True.
    format : str, optional
        Custom log message format. If not provided, a default format will be used.
    clear_handlers : bool, optional
        Whether to clear existing handlers from the logger before adding new ones (default is False).

    Returns
    -------
    logging.Logger
        Configured logger instance.

    """

    stream_level = verbose_to_log_level(stream_level)
    file_level = verbose_to_log_level(file_level)

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    if format is None:
        # format = '[%(asctime)s | %(name)s | %(levelname)s: %(message)s'
        format = '%(asctime)s | %(name)s | %(levelname)7s - %(message)s'

    formatter = logging.Formatter(format, datefmt='%H:%M:%S')
    # formatter = TruncatingFormatter(format, name_max=16, datefmt='%H:%M:%S')

    ## optionally clear existing handlers
    if clear_handlers:
        logger.handlers.clear()

    if stream_logger:
        h = logging.StreamHandler()
        h.setLevel(stream_level)
        h.setFormatter(formatter)
        logger.addHandler(h)

    if file_logger:
        assert out_dir is not None, 'out_dir must be provided if file_logger is True'
        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)

        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        log_path = out_dir / f'{logger_name}_{timestamp}.log'

        prior_size = log_path.stat().st_size if log_path.exists() else 0
        if file_mode == 'w':
            prior_size = 0

        fh = logging.FileHandler(log_path, mode=file_mode, encoding='utf-8')
        fh.setLevel(file_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

        session_lines = ['log session created']
        if prior_size > 0:
            session_lines.insert(0, '')
        fh.stream.write('\n'.join(session_lines) + '\n')
        fh.flush()

    return logger


class LoggerWriter:
    def __init__(self, logger, level):
        self.logger = logger
        self.level = level
        self._buffer = ''

    def write(self, message):
        # stdout writes can be chunked, so we buffer and split on newlines
        if not isinstance(message, str):
            message = str(message)

        self._buffer += message
        while '\n' in self._buffer:
            line, self._buffer = self._buffer.split('\n', 1)
            line = line.strip()
            if line:
                self.logger.log(self.level, line)

    def flush(self):
        # Flush any remaining text in the buffer
        if self._buffer.strip():
            self.logger.log(self.level, self._buffer.strip())
        self._buffer = ''


def symlink_original_files(g4x_obj, out_dir: Path | str) -> None:
    ignore_file_list = ['clustering_umap.csv.gz', 'dgex.csv.gz', 'transcript_panel.csv']
    data_dir = Path(g4x_obj.data_dir).resolve()

    for root, dirs, files in os.walk(data_dir):
        rel_root = Path(root).relative_to(data_dir)
        if str(rel_root) == 'metrics':
            continue
        dst_root = out_dir / rel_root
        dst_root.mkdir(parents=True, exist_ok=True)

        for f in files:
            if f in ignore_file_list:
                continue
            # src_file = Path(root) / f
            dst_file = dst_root / f
            if dst_file.exists():
                dst_file.unlink()
            dst_file.symlink_to((Path(root) / f).resolve())


def parse_input_manifest(file_path: Path, verbose: bool = False) -> pl.DataFrame:
    """
    Parse strings in `df[col]` of the form '<prefix>-<15mer>-<primer>'
    and add columns: gene_name, sequence, primer, primer_code, read.
    Rows that don't match the pattern get nulls in the new columns.
    """
    # print(f'Parsing transcript manifest: {file_path.name}')
    df = pl.read_csv(file_path)

    col = 'probe_name'
    if col not in df.columns:
        raise ValueError(f"transcript manifest must contain column '{col}'")

    parsed = df.with_columns(
        [
            pl.col(col).str.extract(PROBE_PATTERN, 1).alias('gene_name'),
            pl.col(col).str.extract(PROBE_PATTERN, 2).alias('sequence'),
            pl.col(col).str.extract(PROBE_PATTERN, 3).alias('primer'),
        ]
    )

    null_count = parsed.null_count()['sequence'][0]
    if null_count > 0:
        if verbose:
            print(f'{null_count} probes with invalid sequence format will be ignored:')
            null_seqs = parsed.filter(pl.col('sequence').is_null())['probe_name'].to_list()
            for ns in null_seqs:
                print(f'- {ns}')

    if 'panel_type' in df.columns:
        parsed = parsed.drop('panel_type')

    if 'read' in df.columns:
        if verbose:
            print("Using 'read' column provided by input manifest.")
        parsed = parsed.drop('read')
    else:
        plist = parsed['primer'].unique().to_list()
        ign_primer = [p for p in plist if p not in primer_read_map]
        if len(ign_primer) > 0:
            if verbose:
                print('Warning: the following primer names are not known and will be ignored:')
                for ip in ign_primer:
                    print(f'- {ip}')
        parsed = parsed.filter(pl.col('primer').is_in(primer_read_map.keys())).with_columns(
            pl.col('primer').replace(primer_read_map).cast(pl.Int8).alias('read')
        )

    if 'gene_name' in df.columns:
        if verbose:
            print("Using 'gene_name' column provided by input manifest.")
        parsed = parsed.drop('gene_name')

    manifest = parsed.join(df, on=col, how='left')

    manifest = manifest.with_columns(
        probe_type=(
            pl.when(pl.col('gene_name').str.to_lowercase() == 'gdna')
            .then(pl.lit('GCP'))
            .when(pl.col('gene_name').str.starts_with('NCS-'))
            .then(pl.lit('NCS'))
            .when(pl.col('gene_name').str.starts_with('NCP-'))
            .then(pl.lit('NCP'))
            .otherwise(pl.lit('targeting'))
        )
    )

    return manifest



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: CellMasksSchema.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15\x43\x65llMasksSchema.proto")\n\tUmapEntry\x12\r\n\x05umapY\x18\x01 \x01(\x01\x12\r\n\x05umapX\x18\x02 \x01(\x01"\xea\x01\n\nSingleMask\x12\x14\n\x08vertices\x18\x01 \x03(\rB\x02\x10\x01\x12\x0c\n\x04\x61rea\x18\x02 \x01(\r\x12\x13\n\x0btotalCounts\x18\x03 \x01(\r\x12\x12\n\ntotalGenes\x18\x04 \x01(\r\x12\x0e\n\x06\x63\x65llId\x18\x05 \x01(\t\x12\x11\n\tclusterId\x18\x06 \x01(\t\x12\x15\n\rproteinValues\x18\x07 \x03(\r\x12\x1a\n\x12nonzeroGeneIndices\x18\x08 \x03(\r\x12\x19\n\x11nonzeroGeneValues\x18\t \x03(\r\x12\x1e\n\numapValues\x18\n \x01(\x0b\x32\n.UmapEntry"5\n\rColormapEntry\x12\x11\n\tclusterId\x18\x01 \x01(\t\x12\x11\n\x05\x63olor\x18\x02 \x03(\rB\x02\x10\x01"3\n\x08Metadata\x12\x14\n\x0cproteinNames\x18\x01 \x03(\t\x12\x11\n\tgeneNames\x18\x02 \x03(\t"\x81\x01\n\tCellMasks\x12\x1e\n\tcellMasks\x18\x01 \x03(\x0b\x32\x0b.SingleMask\x12 \n\x08\x63olormap\x18\x02 \x03(\x0b\x32\x0e.ColormapEntry\x12\x1b\n\x08metadata\x18\x03 \x01(\x0b\x32\t.Metadata\x12\x15\n\rnumberOfCells\x18\x04 \x01(\rb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'CellMasksSchema_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_SINGLEMASK'].fields_by_name['vertices']._loaded_options = None
    _globals['_SINGLEMASK'].fields_by_name['vertices']._serialized_options = b'\020\001'
    _globals['_COLORMAPENTRY'].fields_by_name['color']._loaded_options = None
    _globals['_COLORMAPENTRY'].fields_by_name['color']._serialized_options = b'\020\001'
    _globals['_UMAPENTRY']._serialized_start = 25
    _globals['_UMAPENTRY']._serialized_end = 66
    _globals['_SINGLEMASK']._serialized_start = 69
    _globals['_SINGLEMASK']._serialized_end = 303
    _globals['_COLORMAPENTRY']._serialized_start = 305
    _globals['_COLORMAPENTRY']._serialized_end = 358
    _globals['_METADATA']._serialized_start = 360
    _globals['_METADATA']._serialized_end = 411
    _globals['_CELLMASKS']._serialized_start = 414
    _globals['_CELLMASKS']._serialized_end = 543
# @@protoc_insertion_point(module_scope)



import traceback
from contextlib import contextmanager

import rich_click as click
from rich.console import Console

from .. import utils

console = Console()


# click.rich_click.THEME = 'modern'
click.rich_click.MAX_WIDTH = 100
click.rich_click.COMMANDS_PANEL_TITLE = 'commands'
click.rich_click.OPTIONS_PANEL_TITLE = 'options'
click.rich_click.STYLE_OPTION = 'bold blue'
click.rich_click.STYLE_ARGUMENT = 'bold blue'
click.rich_click.STYLE_COMMAND = 'bold blue'
click.rich_click.STYLE_SWITCH = 'bold red'
click.rich_click.STYLE_METAVAR = 'bold red'
click.rich_click.STYLE_METAVAR_SEPARATOR = 'dim'
click.rich_click.STYLE_USAGE = 'bold yellow'
click.rich_click.STYLE_USAGE_COMMAND = 'bold'
click.rich_click.STYLE_HELPTEXT_FIRST_LINE = ''
click.rich_click.STYLE_HELPTEXT = 'dim'
click.rich_click.STYLE_OPTION_DEFAULT = 'dim'
click.rich_click.STYLE_REQUIRED_SHORT = 'bold yellow'
click.rich_click.STYLE_REQUIRED_LONG = 'bold yellow'
click.rich_click.STYLE_OPTIONS_PANEL_BORDER = 'dim'
click.rich_click.STYLE_COMMANDS_PANEL_BORDER = 'dim'
click.rich_click.COMMANDS_BEFORE_OPTIONS = True

click.rich_click.ARGUMENTS_PANEL_TITLE = 'input'

click.rich_click.COMMAND_GROUPS = {
    'g4x-helpers': [
        {'name': 'commands', 'commands': ['redemux', 'resegment', 'update_bin', 'new_bin', 'tar_viewer']},
        # {"name": "utilities", "commands": ["log"]},
    ],
}


# click.rich_click.OPTION_GROUPS = {
#     'g4x-helpers': [
#         {
#             'name': 'in/out',  #
#             'options': ['--input', '--output'],
#         },
#         {
#             'name': 'options',  #
#             'options': [
#                 '--sample-id',
#                 '--threads',
#                 '--verbose',
#                 '--version',
#                 '--help',
#             ],
#         },
#     ]
# }


@contextmanager
def _spinner(message: str):
    with console.status(message, spinner='dots', spinner_style='red'):
        yield


def _fail_message(func_name, e, trace_back=False):
    click.echo('')
    click.secho(f'Failed {func_name}:', fg='red', err=True, bold=True)
    if trace_back:
        traceback.print_exc()
    raise click.ClickException(f'{type(e).__name__}: {e}')


def initialize_sample(
    data_dir: str, sample_id: str | None = None, in_place: bool = False, n_threads: int = utils.DEFAULT_THREADS
) -> None:
    msg = f'loading G4X-data from [blue]{data_dir}[/blue]'
    with _spinner(msg):
        import glymur

        from ..models import G4Xoutput

        glymur.set_option('lib.num_threads', n_threads)
        try:
            sample = G4Xoutput(data_dir=data_dir, sample_id=sample_id)
        except Exception as e:
            click.echo('\n')
            click.secho('Failed to load G4X-data:', fg='red', err=True, bold=True)
            raise click.ClickException(f'{e}')

    if in_place:
        out_dir = sample.data_dir
        click.secho('Editing in-place!', fg='blue', bold=True)
    else:
        out_dir = sample.data_dir / 'g4x_helpers'

    return sample, out_dir


def print_k_v(item, value, gap=2):
    value = '<undefined>' if not value else value
    click.secho(f'{item:<{gap}}', dim=True, nl=False)
    click.secho('- ', dim=True, nl=False)
    click.secho(f'{value}', fg='blue', bold=True)



import inspect

from .. import __version__, utils
from . import cli_setup
from . import help_messages as hm

click = cli_setup.click


# region cli
@click.group(
    context_settings=dict(help_option_names=['-h', '--help']),
    invoke_without_command=True,
    add_help_option=True,
    help=hm.CLI_HELP,
)
@click.option(
    '-t',
    '--threads',
    required=False,
    type=int,
    default=utils.DEFAULT_THREADS,
    show_default=True,
    help='Number of threads to use for processing',
)
@click.option(
    '-v',
    '--verbose',
    required=False,
    type=int,
    default=1,
    show_default=True,
    help='Console logging level (0, 1, 2)',
)
@click.option(
    '--version',
    is_flag=True,
    default=False,
    help='Display g4x-helpers version',
)
@click.pass_context
# @click.option('-v', '--verbose', default=2, count=True, help='Console logging level (0, 1, 2)')
def cli(ctx, threads, verbose, version):
    if version:
        click.echo(f'g4x-helpers: {__version__}')
        ctx.exit()

    # No subcommand and no input given → show help
    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())
        ctx.exit()

    if ctx.invoked_subcommand:
        ctx.ensure_object(dict)

        ctx.obj['threads'] = threads
        ctx.obj['verbose'] = verbose
        ctx.obj['version'] = __version__


def g4x_data_opt():
    return click.argument(
        'g4x-data',
        type=click.Path(exists=True, file_okay=False),
        help='Directory containing G4X-data for a single sample',
        # panel='data i/o',
    )


def in_place_opt(cmd_name: str = ''):
    return click.option(
        '-ip',
        '--in-place',
        is_flag=True,
        help=f'Edit G4X-data in-place if this flag is set.\n\nOtherwise creates a "g4x_helpers/{cmd_name}" folder.',
    )


############################################################
# region resegment
name = 'resegment'


@cli.command(name=name, help=hm.RESEG_HELP)
@g4x_data_opt()
@click.option(
    '--cell-labels',
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='File containing cell segmentation labels.\n\nsupported file types: [.npy, .npz, .geojson]',
)
@click.option(
    '--labels-key',
    required=False,
    type=str,
    default=None,
    help='Key/column in npz/geojson where labels should be taken from (optional, but required for .npz with multiple arrays)',
)
@in_place_opt(name)
@click.pass_context
def resegment(ctx, g4x_data, cell_labels, labels_key, in_place):
    func_name = inspect.currentframe().f_code.co_name

    g4x_obj, out_dir = cli_setup.initialize_sample(data_dir=g4x_data, in_place=in_place, n_threads=ctx.obj['threads'])

    try:
        with cli_setup._spinner(f'Initializing {func_name} process...'):
            from ..main_features import resegment as main_resegment

        main_resegment(
            g4x_obj=g4x_obj,
            out_dir=out_dir,
            cell_labels=cell_labels,
            labels_key=labels_key,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        cli_setup._fail_message(func_name, e)


############################################################
# region redemux
name = 'redemux'


@cli.command(name=name, help=hm.REDMX_HELP)
@g4x_data_opt()
@click.option(
    '--manifest',
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to manifest for demuxing.\n\n Must contain a "probe_name" column with the format "geneid-sequence-primer"',
)
@click.option(
    '--batch-size',
    default=1_000_000,
    show_default=True,
    type=int,
    help='Number of transcripts to process per batch.',
)
@in_place_opt(name)
@click.pass_context
def redemux(ctx, g4x_data, manifest, batch_size, in_place):
    func_name = inspect.currentframe().f_code.co_name
    g4x_obj, out_dir = cli_setup.initialize_sample(data_dir=g4x_data, in_place=in_place, n_threads=ctx.obj['threads'])
    try:
        with cli_setup._spinner(f'Initializing {func_name} process...'):
            from ..main_features import redemux as main_redemux

        main_redemux(
            g4x_obj=g4x_obj,
            out_dir=out_dir,
            manifest=manifest,
            batch_size=batch_size,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        cli_setup._fail_message(func_name, e)


############################################################
# region update_bin
name = 'update_bin'


@cli.command(name=name, help=hm.UDBIN_HELP)
@g4x_data_opt()
@click.option(
    '--metadata',
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to metadata table with clustering and/or embedding information. Must contain cell-IDs that match those in the bin file.',
)
@click.option(
    '--cellid-key',
    default='cell_id',
    type=str,
    help='Column name in metadata containing cell-IDs.\n\n If not provided, looks for column named "cell_id"',
)
@click.option(
    '--cluster-key',
    default=None,
    type=str,
    help='Column name in metadata containing cluster IDs.\n\n If not provided, skips updating cluster IDs.',
)
@click.option(
    '--cluster-color-key',
    default=None,
    type=str,
    help='Column name in metadata containing cluster colors.\n\n (format: hex) Only active if cluster_key is updated.\n\n If not provided, assigns colors automatically.',
)
@click.option(
    '--emb-key',
    default=None,
    type=str,
    help='Column name in metadata containing 2D-embedding coordinates.\n\n Parser will look for {emb_key}_1 and {emb_key}_2.\n\n If not provided, skips updating embedding.',
)
@in_place_opt(name)
@click.pass_context
def update_bin(ctx, g4x_data, metadata, cellid_key, cluster_key, cluster_color_key, emb_key, in_place):
    func_name = inspect.currentframe().f_code.co_name
    g4x_obj, out_dir = cli_setup.initialize_sample(data_dir=g4x_data, in_place=in_place, n_threads=ctx.obj['threads'])
    try:
        with cli_setup._spinner(f'Initializing {func_name} process...'):
            from ..main_features import update_bin as main_update_bin

        main_update_bin(
            g4x_obj=g4x_obj,
            bin_file=g4x_obj.data_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_segmentation.bin',
            bin_out=out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_segmentation.bin',
            out_dir=out_dir,
            metadata=metadata,
            cellid_key=cellid_key,
            cluster_key=cluster_key,
            cluster_color_key=cluster_color_key,
            emb_key=emb_key,
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        cli_setup._fail_message(func_name, e)


############################################################
# region new_bin
name = 'new_bin'


@cli.command(name=name, help=hm.NWBIN_HELP)
@g4x_data_opt()
@in_place_opt(name)
@click.pass_context
def new_bin(ctx, g4x_data, in_place):
    func_name = inspect.currentframe().f_code.co_name
    g4x_obj, out_dir = cli_setup.initialize_sample(data_dir=g4x_data, in_place=in_place, n_threads=ctx.obj['threads'])
    try:
        with cli_setup._spinner(f'Initializing {func_name} process...'):
            from ..main_features import new_bin as main_new_bin

        main_new_bin(
            g4x_obj=g4x_obj,
            out_dir=out_dir,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        cli_setup._fail_message(func_name, e)


############################################################
# region tar_viewer
name = 'tar_viewer'


@cli.command(name=name, help=hm.TARVW_HELP)
@g4x_data_opt()
@in_place_opt(name)
@click.pass_context
def tar_viewer(ctx, g4x_data, in_place):
    func_name = inspect.currentframe().f_code.co_name

    g4x_obj, out_dir = cli_setup.initialize_sample(data_dir=g4x_data, in_place=in_place, n_threads=ctx.obj['threads'])
    try:
        with cli_setup._spinner(f'Initializing {func_name} process...'):
            from ..main_features import tar_viewer as main_tar_viewer

        main_tar_viewer(
            g4x_obj=g4x_obj,
            out_dir=out_dir,
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        cli_setup._fail_message(func_name, e)


############################################################
# region migrate
@cli.command(name='migrate', help=hm.MIGRT_HELP)
@g4x_data_opt()
@click.option(
    '--restore',
    is_flag=True,
    help='Restores an existing migration-backup.',
)
@click.pass_context
def migrate(ctx, g4x_data, restore):
    func_name = inspect.currentframe().f_code.co_name

    g4x_obj, _ = cli_setup.initialize_sample(data_dir=g4x_data, in_place=False, n_threads=ctx.obj['threads'])

    try:
        with cli_setup._spinner(f'Initializing {func_name} process...'):
            from ..main_features import migrate as main_migrate

        main_migrate(
            g4x_obj=g4x_obj,
            restore=restore,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        cli_setup._fail_message(func_name, e)


############################################################
# region validate
@cli.command(name='validate', help=hm.VLDTE_HELP)
@g4x_data_opt()
@click.pass_context
def validate(ctx, g4x_data):
    func_name = inspect.currentframe().f_code.co_name

    g4x_obj, _ = cli_setup.initialize_sample(data_dir=g4x_data, in_place=False, n_threads=ctx.obj['threads'])

    try:
        with cli_setup._spinner(f'Initializing {func_name} process...'):
            from ..main_features import validate as main_validate

        main_validate(
            g4x_obj=g4x_obj,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        cli_setup._fail_message(func_name, e)


if __name__ == '__main__':
    cli(prog_name='g4x-helpers')



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class UmapEntry(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    UMAPY_FIELD_NUMBER: builtins.int
    UMAPX_FIELD_NUMBER: builtins.int
    umapY: builtins.float
    umapX: builtins.float
    def __init__(
        self,
        *,
        umapY: builtins.float = ...,
        umapX: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['umapX', b'umapX', 'umapY', b'umapY']) -> None: ...

global___UmapEntry = UmapEntry

@typing.final
class SingleMask(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VERTICES_FIELD_NUMBER: builtins.int
    AREA_FIELD_NUMBER: builtins.int
    TOTALCOUNTS_FIELD_NUMBER: builtins.int
    TOTALGENES_FIELD_NUMBER: builtins.int
    CELLID_FIELD_NUMBER: builtins.int
    CLUSTERID_FIELD_NUMBER: builtins.int
    PROTEINVALUES_FIELD_NUMBER: builtins.int
    NONZEROGENEINDICES_FIELD_NUMBER: builtins.int
    NONZEROGENEVALUES_FIELD_NUMBER: builtins.int
    UMAPVALUES_FIELD_NUMBER: builtins.int
    area: builtins.int
    totalCounts: builtins.int
    totalGenes: builtins.int
    cellId: builtins.str
    clusterId: builtins.str
    @property
    def vertices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def proteinValues(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def nonzeroGeneIndices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def nonzeroGeneValues(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def umapValues(self) -> global___UmapEntry: ...
    def __init__(
        self,
        *,
        vertices: collections.abc.Iterable[builtins.int] | None = ...,
        area: builtins.int = ...,
        totalCounts: builtins.int = ...,
        totalGenes: builtins.int = ...,
        cellId: builtins.str = ...,
        clusterId: builtins.str = ...,
        proteinValues: collections.abc.Iterable[builtins.int] | None = ...,
        nonzeroGeneIndices: collections.abc.Iterable[builtins.int] | None = ...,
        nonzeroGeneValues: collections.abc.Iterable[builtins.int] | None = ...,
        umapValues: global___UmapEntry | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal['umapValues', b'umapValues']) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            'area',
            b'area',
            'cellId',
            b'cellId',
            'clusterId',
            b'clusterId',
            'nonzeroGeneIndices',
            b'nonzeroGeneIndices',
            'nonzeroGeneValues',
            b'nonzeroGeneValues',
            'proteinValues',
            b'proteinValues',
            'totalCounts',
            b'totalCounts',
            'totalGenes',
            b'totalGenes',
            'umapValues',
            b'umapValues',
            'vertices',
            b'vertices',
        ],
    ) -> None: ...

global___SingleMask = SingleMask

@typing.final
class ColormapEntry(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CLUSTERID_FIELD_NUMBER: builtins.int
    COLOR_FIELD_NUMBER: builtins.int
    clusterId: builtins.str
    @property
    def color(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        clusterId: builtins.str = ...,
        color: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['clusterId', b'clusterId', 'color', b'color']) -> None: ...

global___ColormapEntry = ColormapEntry

@typing.final
class Metadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROTEINNAMES_FIELD_NUMBER: builtins.int
    GENENAMES_FIELD_NUMBER: builtins.int
    @property
    def proteinNames(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def geneNames(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        proteinNames: collections.abc.Iterable[builtins.str] | None = ...,
        geneNames: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal['geneNames', b'geneNames', 'proteinNames', b'proteinNames']
    ) -> None: ...

global___Metadata = Metadata

@typing.final
class CellMasks(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CELLMASKS_FIELD_NUMBER: builtins.int
    COLORMAP_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    NUMBEROFCELLS_FIELD_NUMBER: builtins.int
    numberOfCells: builtins.int
    @property
    def cellMasks(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SingleMask]: ...
    @property
    def colormap(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ColormapEntry]: ...
    @property
    def metadata(self) -> global___Metadata: ...
    def __init__(
        self,
        *,
        cellMasks: collections.abc.Iterable[global___SingleMask] | None = ...,
        colormap: collections.abc.Iterable[global___ColormapEntry] | None = ...,
        metadata: global___Metadata | None = ...,
        numberOfCells: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal['metadata', b'metadata']) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            'cellMasks',
            b'cellMasks',
            'colormap',
            b'colormap',
            'metadata',
            b'metadata',
            'numberOfCells',
            b'numberOfCells',
        ],
    ) -> None: ...

global___CellMasks = CellMasks



import functools
import logging
from typing import TYPE_CHECKING

import rich_click as click

from . import __version__, utils
from .cli.cli_setup import print_k_v

if TYPE_CHECKING:
    from .models import G4Xoutput


def _base_command(func):
    """Decorator to apply standard command initialization logic."""

    @functools.wraps(func)
    def wrapper(
        g4x_obj: 'G4Xoutput',
        out_dir: str | None = None,
        n_threads: int = utils.DEFAULT_THREADS,
        verbose: int = 1,
        file_logger: bool = True,
        **kwargs,
    ):
        func_name = func.__name__

        if func_name not in ('migrate', 'validate'):
            g4x_obj.validate()

        out_dir = g4x_obj.data_dir if out_dir is None else out_dir
        out_dir = utils.validate_path(out_dir, must_exist=False, is_dir_ok=True, is_file_ok=False)

        if out_dir != g4x_obj.data_dir:
            func_out = out_dir / func_name
            func_out.mkdir(parents=True, exist_ok=True)
            out_dir = func_out

        gap = 12
        click.secho(f'\nStarting: {func_name}\n', bold=True)
        print_k_v('sample_dir', f'{g4x_obj.data_dir}', gap)
        print_k_v('out_dir', f'{out_dir}', gap)
        print_k_v('n_threads', f'{n_threads}', gap)
        print_k_v('verbosity', f'{verbose}', gap)
        print_k_v('g4x-helpers', f'v{__version__}', gap)
        click.echo('')

        log_dir = g4x_obj.data_dir / 'g4x_helpers' / 'logs'
        log_dir.mkdir(parents=True, exist_ok=True)
        if not kwargs.get('logger', None):
            logger = utils.setup_logger(
                logger_name=func_name, out_dir=log_dir, stream_level=verbose, file_logger=file_logger
            )
        logger.info(f'Running {func_name} with G4X-helpers v{__version__}')

        result = func(
            g4x_obj=g4x_obj,
            out_dir=out_dir,
            n_threads=n_threads,
            verbose=verbose,
            logger=logger,
            **kwargs,
        )

        click.secho(f'\nCompleted: {func.__name__}', bold=True, fg='green')
        return result

    return wrapper


@_base_command
def resegment(
    g4x_obj: 'G4Xoutput',
    out_dir: str,
    cell_labels: str,
    *,
    labels_key: str | None = None,
    n_threads: int = utils.DEFAULT_THREADS,
    logger: logging.Logger,
    **kwargs,
) -> None:
    from .modules import init_bin, segment

    cell_labels = utils.validate_path(cell_labels, must_exist=True, is_dir_ok=False, is_file_ok=True)

    logger.info('Loading segmentation mask.')
    labels = segment.try_load_segmentation(cell_labels=cell_labels, expected_shape=g4x_obj.shape, labels_key=labels_key)

    segment.apply_segmentation(
        g4x_obj=g4x_obj,
        labels=labels,
        out_dir=out_dir,
        skip_protein_extraction=False,
        logger=logger,
    )

    init_bin.init_bin_file(
        g4x_obj=g4x_obj,
        seg_mask=labels,
        single_cell_dir=out_dir / 'single_cell_data',
        out_dir=out_dir,
        logger=logger,
        n_threads=n_threads,
    )

    # TODO tell user to run update_bin afterwards

    # bin_file = out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_segmentation.bin'
    # edit_bin.edit_bin_file(g4x_obj=g4x_obj, bin_file=bin_file, logger=logger)


@_base_command
def redemux(
    g4x_obj: 'G4Xoutput',
    out_dir: str,
    manifest: str,
    *,
    batch_size: int = 1_000_000,
    n_threads: int = utils.DEFAULT_THREADS,
    logger: logging.Logger,
    **kwargs,
) -> None:
    from g4x_helpers.modules import demux, init_bin, segment, transcript_tar

    demux.demux_raw_features(
        g4x_obj=g4x_obj,
        manifest=manifest,
        out_dir=out_dir,
        batch_size=batch_size,
        logger=logger,
    )

    labels = g4x_obj.load_segmentation()

    segment.apply_segmentation(
        g4x_obj=g4x_obj,
        labels=labels,
        tx_table=out_dir / 'rna' / 'transcript_table.csv.gz',
        manifest=manifest,
        out_dir=out_dir,
        skip_protein_extraction=False,
        create_source=False,
        logger=logger,
    )

    init_bin.init_bin_file(
        g4x_obj=g4x_obj,
        single_cell_dir=out_dir / 'single_cell_data',
        out_dir=out_dir,
        n_threads=n_threads,
        seg_mask=labels,
        logger=logger,
    )

    # TODO tell user to run update_bin afterwards

    # bin_file = out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_segmentation.bin'
    # edit_bin.edit_bin_file(g4x_obj=g4x_obj, bin_file=bin_file, logger=logger)

    transcript_tar.create_tx_tarfile(
        g4x_obj=g4x_obj,
        tx_table=out_dir / 'rna' / 'transcript_table.csv.gz',
        out_path=out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_transcripts.tar',
        n_threads=n_threads,
        logger=logger,
    )


@_base_command
def update_bin(
    g4x_obj: 'G4Xoutput',
    bin_file: str,
    metadata: str,
    out_dir: str,
    *,
    cellid_key: str | None = None,
    cluster_key: str | None = None,
    clustercolor_key: str | None = None,
    emb_key: str | None = None,
    logger: logging.Logger,
    **kwargs,
) -> None:
    from .modules import edit_bin

    metadata = utils.validate_path(metadata, must_exist=True, is_dir_ok=False, is_file_ok=True)

    edit_bin.edit_bin_file(
        g4x_obj=g4x_obj,
        bin_file=bin_file,
        metadata=metadata,
        bin_out=out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_segmentation.bin',
        cellid_key=cellid_key,
        cluster_key=cluster_key,
        clustercolor_key=clustercolor_key,
        emb_key=emb_key,
        logger=logger,
    )


@_base_command
def new_bin(
    g4x_obj: 'G4Xoutput',  #
    out_dir: str,
    *,
    n_threads: int = utils.DEFAULT_THREADS,
    logger: logging.Logger,
    **kwargs,
) -> None:
    from .modules import init_bin

    init_bin.init_bin_file(
        g4x_obj=g4x_obj,
        out_dir=out_dir,
        n_threads=n_threads,
        logger=logger,
    )


@_base_command
def tar_viewer(
    g4x_obj: 'G4Xoutput',  #
    out_dir: str | None = None,
    *,
    logger: logging.Logger,
    **kwargs,
) -> None:
    from .modules import viewer_dir

    viewer_dir.package_viewer_dir(
        viewer_dir=g4x_obj.data_dir / 'g4x_viewer',
        sample_id=g4x_obj.sample_id,
        out_dir=out_dir,
        logger=logger,
    )


@_base_command
def migrate(
    g4x_obj: 'G4Xoutput',
    restore: bool = False,
    n_threads: int = utils.DEFAULT_THREADS,
    *,
    logger: logging.Logger,
    **kwargs,
) -> None:
    from .schemas import migration

    if restore:
        migration.restore_backup(data_dir=g4x_obj.data_dir, sample_id=g4x_obj.sample_id, logger=logger)
    else:
        migration.migrate_g4x_data(
            data_dir=g4x_obj.data_dir, sample_id=g4x_obj.sample_id, n_threads=n_threads, logger=logger
        )


@_base_command
def validate(
    g4x_obj: 'G4Xoutput',
    *,
    logger: logging.Logger,
    **kwargs,
):
    from .schemas import validate

    try:
        validate.validate_g4x_data(
            g4x_obj.data_dir, schema_name='base_schema', formats={'sample_id': g4x_obj.sample_id}, report='long'
        )
    except validate.ValidationError as e:
        print('Directory structure validation failed!: ', e)

    try:
        print('\n')
        validate.validate_file_schemas(g4x_obj.data_dir, verbose=True)
    except validate.ValidationError as e:
        print('File schema validation failed!: ', e)



import logging
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.colors as mcolors
import numpy as np
import polars as pl
from tqdm import tqdm

from .. import utils
from ..g4x_viewer import CellMasksSchema_pb2 as CellMasksSchema
from ..schemas.validate import read_bin_file
from .workflow import workflow

if TYPE_CHECKING:
    from ..models import G4Xoutput


DEFAULT_COLOR = '#BFBFBF'

sg_palette = [
    '#72FFAB',
    '#A16CFD',
    '#FF7043',
    '#008FFF',
    '#D32F2F',
    '#7CB342',
    '#7F34BE',
    '#FFCA28',
    '#0C8668',
    '#FB4695',
    '#005EE1',
    '#28EDED',
    '#A17B64',
    '#FFFF58',
    '#BC29AE',
    '#006D8F',
    '#FFBAFF',
    '#FFD091',
    '#5C6BC0',
    '#F490B2',
]


@workflow
def edit_bin_file(
    g4x_obj: 'G4Xoutput',
    bin_file: Path,
    *,
    metadata: Path | None = None,
    cellid_key: str | None = None,
    cluster_key: str | None = None,
    clustercolor_key: str | None = None,
    emb_key: str | None = None,
    bin_out: Path | None = None,
    logger: logging.Logger,
) -> None:
    logger.info(f'Reading bin file: {bin_file}')

    bin_file = utils.validate_path(bin_file, must_exist=True, is_dir_ok=False, is_file_ok=True)

    if not bin_out:
        bin_out = bin_file
    else:
        bin_out = utils.validate_path(bin_out, must_exist=False, is_dir_ok=False, is_file_ok=True)

    cell_masks = read_bin_file(bin_file)

    update_clusters = update_emb = False

    if metadata is None:
        logger.info('No metadata provided. Building from adata source.')

        unused_keys = [k for k, v in locals().items() if k.endswith('_key') and v is not None]
        if len(unused_keys) > 0:
            logger.info(f'Ignoring provided keys: {unused_keys}. Only needed when metadata is provided.')

        logger.info('Loading adata.')
        obs_df = g4x_obj.load_adata(remove_nontargeting=False, load_clustering=True).obs
        obs_df = pl.from_pandas(obs_df)

        if 'cell_id' not in obs_df.columns:
            raise ValueError('Invalid adata. No "cell_id" column found in adata.obs.')

        cellid_key = 'cell_id'
        cluster_key = infer_cluster_key(obs_df)
        emb_key = infer_emb_key(obs_df)
        clustercolor_key = None

        obs_df = obs_df.cast({cluster_key: pl.UInt16})
        update_clusters = cluster_key is not None
        update_emb = emb_key is not None

    if metadata is not None:
        if emb_key is None and cluster_key is None:
            logger.info('neither emb_key nor cluster_key were provided, nothing to update.')
            return None

        obs_df = pl.read_csv(metadata)

        if cellid_key is not None:
            check_key(obs_df, 'cellid_key', cellid_key)
        else:
            logger.info('No cellid_key provided, attempting to infer from "label" column.')
            if 'label' in obs_df.columns:
                cellid_key = 'label'
            else:
                raise KeyError('cellid_key not provided and "label" not found in metadata. please provide cellid_key.')

        if cluster_key is not None:
            logger.info(f'Checking cluster_key: {cluster_key}')
            check_key(obs_df, 'cluster_key', cluster_key)

        if emb_key is not None:
            logger.info(f'Checking emb_key: {emb_key}')
            if f'{emb_key}_1' not in obs_df.columns and f'{emb_key}_2' not in obs_df.columns:
                raise ValueError(
                    f'Matching columns for emb_key: "{emb_key}" missing. \nRequired columns: {emb_key}_1 and {emb_key}_2'
                )

        if clustercolor_key is not None and cluster_key is None:
            raise ValueError('clustercolor_key provided without cluster_key. Please provide cluster_key as well.')
        if clustercolor_key is not None:
            logger.info(f'Checking clustercolor_key: {clustercolor_key}')
            check_key(obs_df, 'clustercolor_key', clustercolor_key)

        update_clusters = cluster_key is not None
        update_emb = emb_key is not None

    logger.info('Preparing metadata for updating.')
    obs_df = obs_df.with_row_index('meta_index')
    bin_cell_ids = [cell.cellId for cell in cell_masks.cellMasks]
    bin_df = pl.Series(name=cellid_key, values=bin_cell_ids).to_frame().with_row_index('bin_index')
    obs_df = obs_df.join(bin_df, on=cellid_key, how='right').sort('meta_index')

    not_present = obs_df.null_count()['meta_index'].item()
    if not_present > 0:
        logger.warning(
            f'{not_present} cells in bin file not found in metadata. These will have missing/default values.'
        )

    logger.info('Selecting final columns and filling missing values.')
    final_cols = {
        'cell_id': (cellid_key, 'unknown_cell'),
        'cluster_id': (cluster_key, '-1'),
        'cluster_color': (clustercolor_key, DEFAULT_COLOR),
        'umap_1': (f'{emb_key}_1', float('nan')),
        'umap_2': (f'{emb_key}_2', float('nan')),
    }

    for new_col, (old_col, fill) in final_cols.items():
        if old_col not in obs_df.columns:
            obs_df = obs_df.with_columns(pl.lit(fill).alias(new_col))
        else:
            if new_col != old_col:
                obs_df = obs_df.rename({old_col: new_col})
            obs_df = obs_df.with_columns(pl.col(new_col).fill_null(fill))

    obs_df = obs_df.select(final_cols.keys())

    if update_clusters:
        logger.info('Building cluster palette.')
        cluster_palette = parse_clustercolor_key(obs_df, clustercolor_key)

    ## Do the actual updating
    logger.info('Updating cells.')
    metadata = obs_df.to_pandas().set_index('cell_id')
    for cell in tqdm(cell_masks.cellMasks, desc='Updating cell data'):
        current_cellid = cell.cellId
        if current_cellid not in metadata.index:
            continue

        if update_clusters:
            cell.clusterId = str(metadata.loc[current_cellid, 'cluster_id'])

        if update_emb:
            cell.umapValues.umapX = metadata.loc[current_cellid, 'umap_1']
            cell.umapValues.umapY = metadata.loc[current_cellid, 'umap_2']

    if update_clusters:
        update_colormap(cell_masks, cluster_palette)

    logger.debug(f'Writing updated bin file --> {bin_out}')
    with open(bin_out, 'wb') as file:
        file.write(cell_masks.SerializeToString())


def check_key(obs_df: pl.DataFrame, key_name: str, key: str = None) -> None:
    if key not in obs_df.columns:
        raise KeyError(f'{key_name} "{key}" not found in metadata columns.')


def parse_clustercolor_key(obs_df: pl.DataFrame, clustercolor_key: str | None) -> dict:
    if clustercolor_key:
        color_mapping = obs_df.select('cluster_id', 'cluster_color').unique('cluster_id').sort('cluster_id')

        if not color_mapping['cluster_color'][0].startswith('#'):
            raise ValueError(f'Colors in column "{clustercolor_key}" must be in hex starting with #.')

        cluster_palette = {}
        for cluster, color in color_mapping.iter_rows():
            cluster_palette[cluster] = hex2rgb(color)

    else:
        cluster_palette = generate_cluster_palette(obs_df['cluster_id'])
    return cluster_palette


def hex2rgb(hex: str) -> list[int, int, int]:
    return [int(x * 255) for x in mcolors.to_rgb(hex)]


def infer_cluster_key(obs_df: pl.DataFrame) -> pl.DataFrame:
    cluster_keys = sorted([x for x in obs_df.columns if 'leiden' in x])
    if len(cluster_keys) == 0:
        print('Failed to infer cluster_key, no clustering information will be added.')
        cluster_key = None
    else:
        cluster_key = cluster_keys[0]
        print(f'Inferred cluster_key: {cluster_key}')

    return cluster_key


def infer_emb_key(obs_df: pl.DataFrame) -> pl.DataFrame:
    emb_keys = [c[:-2] for c in obs_df.columns if c.startswith('X_umap')]

    if len(emb_keys) == 0:
        print('No embedding keys available, UMAP coordinates will be missing.')
        emb_key = None
    else:
        emb_key = emb_keys[0]
        print(f'Inferred embedding_key: {emb_key}')
    return emb_key


def generate_cluster_palette(clusters: list, max_colors: int = 256) -> dict:
    """
    Generate a color palette mapping for cluster labels.

    This function assigns RGB colors to unique cluster labels using a matplotlib colormap.
    Clusters labeled as "-1" are assigned a default gray color `[191, 191, 191]`.

    The colormap used depends on the number of clusters:
        - `tab10` for ≤10 clusters
        - `tab20` for ≤20 clusters
        - `hsv` for more than 20 clusters, capped by `max_colors`

    Parameters
    ----------
    clusters : list
        A list of cluster identifiers (strings or integers). The special label '-1' is excluded
        from color mapping and handled separately.
    max_colors : int, optional
        Maximum number of colors to use in the HSV colormap. Only used if there are more than
        20 unique clusters. Default is 256.

    Returns
    -------
    dict
        A dictionary mapping each cluster ID (as a string) to a list of three integers
        representing an RGB color in the range [0, 255].

    Examples
    --------
    >>> generate_cluster_palette(['0', '1', '2', '-1'])
    {'0': [31, 119, 180], '1': [255, 127, 14], '2': [44, 160, 44], '-1': [191, 191, 191]}
    """

    unique_clusters = [c for c in np.unique(clusters) if c != '-1']
    n_clusters = len(unique_clusters)

    if n_clusters <= 20:
        hex_list = sg_palette

    else:
        from matplotlib.pyplot import get_cmap

        hex_list = get_cmap('hsv', min(max_colors, n_clusters)).colors

    cluster_palette = {}
    for i, cluster in enumerate(unique_clusters):
        cluster_palette[str(cluster)] = hex2rgb(hex_list[i])

    cluster_palette['-1'] = hex2rgb(DEFAULT_COLOR)

    return cluster_palette


def update_colormap(cell_masks: CellMasksSchema.CellMasks, cluster_palette: dict[str, list[int]]) -> None:
    cell_masks.ClearField('colormap')
    for cluster_id, color in cluster_palette.items():
        entry = CellMasksSchema.ColormapEntry()
        entry.clusterId = cluster_id
        entry.color.extend(color)
        cell_masks.colormap.append(entry)



from importlib import import_module
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version('g4x_helpers')
except PackageNotFoundError:
    __version__ = 'unknown'

_LAZY_ATTRS = {
    'new_bin': ('.main_features', 'new_bin'),
    'redemux': ('.main_features', 'redemux'),
    'resegment': ('.main_features', 'resegment'),
    'tar_viewer': ('.main_features', 'tar_viewer'),
    'update_bin': ('.main_features', 'update_bin'),
    'migrate': ('.main_features', 'migrate'),
    'G4Xoutput': ('.models', 'G4Xoutput'),
}

__all__ = ['__version__', *sorted(_LAZY_ATTRS)]


def __getattr__(name):
    if name not in _LAZY_ATTRS:
        raise AttributeError(f'module {__name__!r} has no attribute {name!r}')

    module_name, attr_name = _LAZY_ATTRS[name]
    module = import_module(module_name, __name__)
    value = getattr(module, attr_name)
    globals()[name] = value
    return value


def __dir__():
    return sorted(globals().keys() | set(__all__))



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import geopandas
import numpy as np
import pandas as pd
import polars as pl
import scanpy as sc
import shapely.affinity
import skimage.measure
from anndata import AnnData
from geopandas import GeoDataFrame
from rasterio.features import rasterize
from rasterio.transform import Affine
from shapely.geometry import Polygon
from skimage.measure._regionprops import RegionProperties
from tqdm import tqdm

from .. import utils
from .workflow import OutSchema, workflow

if TYPE_CHECKING:
    from geopandas.geodataframe import GeoDataFrame

    from g4x_helpers.models import G4Xoutput


@workflow
def apply_segmentation(
    g4x_obj: 'G4Xoutput',
    labels: np.ndarray,
    out_dir: Path,
    *,
    tx_table: Path | None = None,
    manifest: Path | None = None,
    skip_protein_extraction: bool = False,
    signal_list: list[str] | None = None,
    logger: logging.Logger,
    **kwargs,
):
    out_tree = OutSchema(out_dir, subdirs=['single_cell_data', 'rna', 'segmentation'])

    cell_x_gene_out = out_tree.single_cell_data / 'cell_by_transcript.csv'
    cell_x_protein_out = out_tree.single_cell_data / 'cell_by_protein.csv'
    cell_metadata_out = out_tree.single_cell_data / 'cell_metadata.csv'
    feature_matrix_out = out_tree.single_cell_data / 'feature_matrix.h5'
    tx_table_out = out_tree.rna / 'transcript_table.csv'
    seg_mask_out = out_tree.segmentation / 'segmentation_mask.npz'

    create_source = kwargs.get('create_source', False)
    nuc_mask = g4x_obj.load_segmentation(expanded=False) if create_source else None

    logger.info('Creating cell_by_transcript table.')
    if tx_table is None:
        tx_table = g4x_obj.transcript_table_path
    cell_x_gene, tx_table = create_cell_x_gene(
        g4x_obj=g4x_obj, tx_table=tx_table, manifest=manifest, mask=labels, nucleus_mask=nuc_mask
    )

    logger.info(f'Writing cell_by_transcript matrix to: {cell_x_gene_out.relative_to(out_dir)}.')
    utils.write_csv_gz(df=cell_x_gene, path=cell_x_gene_out)

    logger.info(f'Writing transcript table after adding cell-ids to: {tx_table_out.relative_to(out_dir)}.')
    utils.write_csv_gz(df=tx_table, path=tx_table_out)

    if signal_list is None:
        if g4x_obj.includes_protein:
            signal_list = ['nuclear', 'eosin'] + g4x_obj.proteins
        else:
            signal_list = ['nuclear', 'eosin']
        logger.info(f'Processing image channels: {signal_list}.')
    else:
        logger.info('Processing all image channels.')

    if skip_protein_extraction and g4x_obj.includes_protein:
        logger.warning('Skipping protein extraction. Will try existing cell_by_protein table.')

        cxp_path = g4x_obj.data_dir / 'single_cell_data' / 'cell_by_protein.csv.gz'
        cxp_short = Path(cxp_path.parent.name) / cxp_path.name
        print(f'Cell_x_protein table not provided, loading from: {cxp_short}')
        cell_x_protein = pl.read_csv(cxp_path)

        if signal_list is not None:
            logger.info(f'Selecting channels: {signal_list}')
            signal_list = ['nuclear', 'eosin'] + g4x_obj.proteins
            cell_x_protein = cell_x_protein.select(
                ['label'] + [f'{signal_name}_intensity_mean' for signal_name in signal_list]
            )
    else:
        logger.info('Gathering image intensities.')

        cell_x_protein = create_cell_x_protein(
            g4x_obj=g4x_obj,
            mask=labels,
            signal_list=signal_list,
            cached=False,
        )

        if g4x_obj.includes_protein:
            logger.info(f'Writing cell_by_protein matrix to: {cell_x_protein_out.relative_to(out_dir)}.')
            utils.write_csv_gz(df=cell_x_protein, path=cell_x_protein_out)

    logger.info('Creating adata and metadata.')
    cell_metadata_pre = create_cell_metadata(
        g4x_obj, cell_x_protein=cell_x_protein, mask=labels, create_source=create_source
    )

    adata = create_adata(
        g4x_obj=g4x_obj,
        cell_x_gene=cell_x_gene,
        cell_metadata=cell_metadata_pre,
        manifest=manifest if manifest is not None else g4x_obj.data_dir / 'transcript_panel.csv',
    )

    adata.obs.index = pd.Index(np.asarray(adata.obs.index, dtype=object), dtype=object)
    adata.var.index = pd.Index(np.asarray(adata.var.index, dtype=object), dtype=object)
    logger.info(f'Writing adata to: {feature_matrix_out.relative_to(out_dir)}.')
    adata.write_h5ad(feature_matrix_out)

    logger.info(f'Writing cell metadata table to: {cell_metadata_out.relative_to(out_dir)}.')
    cell_metadata = adata.obs.rename(columns={'cell_id': 'label'}).set_index('label')
    utils.write_csv_gz(df=cell_metadata, path=cell_metadata_out)

    logger.info(f'Writing segmentation mask to: {seg_mask_out.relative_to(out_dir)}.')
    np.savez(seg_mask_out, cell_labels=labels)


def try_load_segmentation(cell_labels: str, expected_shape: tuple[int], labels_key: str | None = None) -> np.ndarray:
    SUPPORTED_MASK_FILETYPES = {'.npy', '.npz', '.geojson'}
    ## load new segmentation
    cell_labels = utils.validate_path(cell_labels, must_exist=True, is_dir_ok=False, is_file_ok=True)
    suffix = cell_labels.suffix.lower()

    if suffix not in SUPPORTED_MASK_FILETYPES:
        raise ValueError(f'{suffix} is not a supported file type.')

    if suffix == '.npz':
        with np.load(cell_labels) as labels:
            available_keys = list(labels.keys())

            if labels_key:  # if a key is specified
                if labels_key not in labels:
                    raise KeyError(f"Key '{labels_key}' not found in .npz; available keys: {available_keys}")
                seg = labels[labels_key]

            else:
                if len(available_keys) != 1:
                    raise ValueError(
                        f'Expected exactly one key in .npz but found {len(available_keys)}: {available_keys}. '
                        "Please specify a key using 'labels_key'."
                    )
                seg = labels[available_keys[0]]

    elif suffix == '.npy':
        # .npy: directly returns the array, no context manager available
        if labels_key is not None:
            print('file is .npy, ignoring provided labels_key.')
        seg = np.load(cell_labels, allow_pickle=False)

    elif suffix == '.geojson':
        gdf = geopandas.read_file(cell_labels)

        if labels_key is not None:
            if labels_key not in gdf.columns:
                raise KeyError(f"Column '{labels_key}' not found in GeoJSON; available columns: {gdf.columns.tolist()}")

            # ensure that a column named 'label' exists
            gdf['label'] = gdf[labels_key]

        else:
            if 'label' not in gdf.columns:
                raise ValueError(
                    "No column named 'label' found in GeoJSON. Please specify which column to use for labels via labels_key."
                )

        print('Rasterizing provided GeoDataFrame.')
        seg = rasterize_polygons(gdf=gdf, target_shape=expected_shape)

    # validate shape for final numpy arrays
    if seg.shape != expected_shape:
        raise ValueError(f'provided mask shape {seg.shape} does not match G4X sample shape {expected_shape}')

    return seg


def get_cell_ids(sample_id: str, mask) -> pl.DataFrame:
    print('Getting cell-IDs from segmentation mask.')
    seg_ids = np.unique(mask[mask != 0])
    cell_ids = (
        pl.Series(name='segmentation_label', values=seg_ids[seg_ids != 0])
        .to_frame()
        .with_columns((f'{sample_id}-' + pl.col('segmentation_label').cast(pl.String)).alias('label'))
        .select(['label', 'segmentation_label'])
    )
    return cell_ids


def get_mask_properties(cell_ids: pl.DataFrame, mask: np.ndarray) -> pl.DataFrame:
    print('Getting regionprops.')
    props = skimage.measure.regionprops(mask)

    prop_dict = []
    # Loop through each region to get the area and centroid, with a progress bar
    for prop in tqdm(props, desc='Extracting mask properties'):
        label = prop.label  # The label (mask id)
        area = prop.area  # Area: count of pixels
        centroid = prop.centroid  # Centroid: (row, col)

        # assuming coordinate order: 'yx':
        cell_y, cell_x = centroid

        px_to_um_area = 0.3125**2

        prop_dict.append(
            {
                'segmentation_label': label,
                'area_um': area * px_to_um_area,
                'cell_x': cell_x,
                'cell_y': cell_y,
            }
        )
    schema = {
        'segmentation_label': pl.Int32,
        'area_um': pl.Float32,
        'cell_x': pl.Float32,
        'cell_y': pl.Float32,
    }
    prop_dict_df = pl.DataFrame(prop_dict, schema=schema)

    prop_dict_df = cell_ids.join(prop_dict_df, on='segmentation_label', how='left')
    return prop_dict_df


def create_cell_metadata(
    g4x_obj: G4Xoutput,
    cell_x_protein: pl.DataFrame | None = None,
    mask: np.ndarray | None = None,
    create_source: bool = False,
) -> pl.DataFrame:
    if create_source:
        cell_props = build_g4x_cell_properties(g4x_obj)
    else:
        cell_props = build_custom_cell_properties(g4x_obj, mask)

    if cell_x_protein is None:
        cxp_path = g4x_obj.data_dir / 'single_cell_data' / 'cell_by_protein.csv.gz'
        cxp_short = Path(cxp_path.parent.name) / cxp_path.name
        print(f'Cell_x_protein table not provided, loading from: {cxp_short}')
        cell_x_protein = pl.read_csv(cxp_path)
        mergeable = cell_x_protein['label'].sort().equals(cell_props['label'].sort())
        if not mergeable:
            raise ValueError(
                f'Cell labels in {cxp_path} do not match those in cell properties. Please provide cell_x_protein DataFrame directly.'
            )

    cell_metadata = cell_props.join(cell_x_protein, on='label')
    return cell_metadata


def build_g4x_cell_properties(g4x_obj: G4Xoutput) -> pl.DataFrame:
    mask = g4x_obj.load_segmentation(expanded=True)
    cell_ids = get_cell_ids(g4x_obj.sample_id, mask)
    cell_props_expanded = get_mask_properties(cell_ids, mask)

    mask = g4x_obj.load_segmentation(expanded=False)
    cell_props_nuc = get_mask_properties(cell_ids, mask)
    del mask

    df = cell_props_nuc.rename({'area_um': 'nuclei_area_um'})
    df_exp = cell_props_expanded.rename({'area_um': 'nuclei_expanded_area_um'}).drop(
        ['segmentation_label', 'cell_x', 'cell_y']
    )

    segmentation_props = df.join(df_exp, on='label').select(
        ['label', 'segmentation_label', 'cell_x', 'cell_y', 'nuclei_area_um', 'nuclei_expanded_area_um']
    )

    return segmentation_props


def build_custom_cell_properties(g4x_obj: G4Xoutput, mask: np.ndarray) -> pl.DataFrame:
    cell_ids = get_cell_ids(g4x_obj.sample_id, mask)
    cell_props = get_mask_properties(cell_ids, mask)
    del mask

    segmentation_props = cell_props.select(['label', 'segmentation_label', 'cell_x', 'cell_y', 'area_um'])

    return segmentation_props


def create_cell_x_gene(
    g4x_obj: G4Xoutput,
    tx_table: Path,
    manifest: Path | None = None,
    *,
    mask: np.ndarray,
    nucleus_mask: np.ndarray | None = None,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    reads = pl.scan_csv(tx_table)

    print('Assigning transcripts to segmentation labels.')
    coord_order = ['y_pixel_coordinate', 'x_pixel_coordinate']
    tx_coords = reads.select(coord_order).collect().to_numpy().astype(int)
    cell_ids = mask[tx_coords[:, 0], tx_coords[:, 1]]

    if nucleus_mask is not None:
        print('Assigning transcripts to nucleus labels.')
        nuc_ids = nucleus_mask[tx_coords[:, 0], tx_coords[:, 1]]
        reads = reads.with_columns(pl.lit(nuc_ids).alias('in_nucleus'))

    if nucleus_mask is None:
        if 'in_nucleus' in reads.collect_schema().names():
            reads = reads.drop('in_nucleus')

    reads = reads.with_columns(pl.lit(cell_ids).alias('segmentation_label')).with_columns(
        cell_id=(g4x_obj.sample_id + '-' + pl.col('segmentation_label').cast(pl.Utf8))
    )

    print('Creating cell x gene matrix.')
    cell_by_gene = (
        reads.filter(pl.col('segmentation_label') != 0)
        .group_by('segmentation_label', 'gene_name')
        .agg(pl.len().alias('counts'))
        .sort('gene_name')
        .collect()
        .pivot(on='gene_name', values='counts', index='segmentation_label')
    )

    print('Adding missing cells with zero counts.')
    cell_ids = get_cell_ids(g4x_obj.sample_id, mask)
    cell_by_gene = (
        cell_ids.join(cell_by_gene, on='segmentation_label', how='left')  # .select('segmentation_label')
        .fill_null(0)
        .drop('segmentation_label')
    )

    print('Adding missing genes with zero counts.')
    zero_count_probes = (
        utils.parse_input_manifest(
            file_path=manifest if manifest is not None else g4x_obj.data_dir / 'transcript_panel.csv'
        )
        .unique('gene_name')
        .filter(~pl.col('gene_name').is_in(cell_by_gene.columns))['gene_name']
        .to_list()
    )

    for probe in zero_count_probes:
        cell_by_gene = cell_by_gene.with_columns(pl.lit(None, dtype=pl.UInt32).alias(probe))

    ordered = [cell_by_gene.columns[0]] + sorted(cell_by_gene.columns[1:])
    cell_by_gene = cell_by_gene.select(ordered).fill_null(0)

    return cell_by_gene, reads.drop('segmentation_label')


def create_cell_x_protein(
    g4x_obj: G4Xoutput,
    mask: np.ndarray,
    signal_list: list[str] | None = None,
    cached: bool = False,
) -> pl.DataFrame | pl.LazyFrame:
    # if signal_list is None:
    #     signal_list = ['nuclear', 'eosin'] + g4x_obj.proteins

    print(f'Creating cell x protein matrix for {len(signal_list)} signals.')

    channel_name_map = {protein: protein for protein in signal_list}
    channel_name_map['nuclear'] = 'nuclearstain'
    channel_name_map['eosin'] = 'cytoplasmicstain'

    # TODO return here when bead masking is implemented
    # bead_mask = g4x_obj.load_bead_mask()
    # bead_mask_flat = bead_mask.ravel() if bead_mask is not None else None
    mask_flat = mask.ravel()

    cell_x_protein = get_cell_ids(g4x_obj.sample_id, mask).drop('segmentation_label')

    for signal_name in tqdm(signal_list, desc='Extracting protein signal'):
        if signal_name not in ['nuclear', 'eosin']:
            image_type = 'protein'
            protein = signal_name
        else:
            image_type = signal_name
            protein = None

        signal_img = g4x_obj.load_image_by_type(image_type, thumbnail=False, protein=protein, cached=cached)

        ch_label = f'{channel_name_map[signal_name]}_intensity_mean'

        intensities = image_mask_intensity_extraction_v2(
            signal_img,
            mask_flat=mask_flat,
            bead_mask_flat=None,
        )

        cell_x_protein = cell_x_protein.with_columns(pl.Series(name=ch_label, values=intensities))

    return cell_x_protein


def create_adata(
    g4x_obj: G4Xoutput,
    cell_x_gene: pl.DataFrame,
    cell_metadata: pl.DataFrame,
    manifest: Path | None = None,
) -> AnnData:
    X = cell_x_gene.drop('label').to_numpy().astype(np.uint16)

    obs_df = cell_metadata.to_pandas()  # .set_index('label')
    obs_df.index = obs_df.index.astype(str)

    gene_ids = pl.Series(name='gene_id', values=cell_x_gene.columns[1:])
    var_df = pl.DataFrame(gene_ids).with_columns(pl.lit('tx').alias('modality'))

    panel_type = (
        utils.parse_input_manifest(manifest if manifest is not None else g4x_obj.data_dir / 'transcript_panel.csv')
        .unique('gene_name')
        .select('gene_name', 'probe_type')
        .rename({'gene_name': 'gene_id'})
    )

    var_df = var_df.join(
        panel_type,
        on='gene_id',
        how='left',
    ).to_pandas()
    var_df.index = var_df.index.astype(str)

    adata = AnnData(X=X, obs=obs_df, var=var_df)
    adata.obs = adata.obs.rename(columns={'label': 'cell_id'})  # if you still want this

    sc.pp.calculate_qc_metrics(adata, inplace=True, percent_top=None)

    return adata


# def image_mask_intensity_extraction(
#     img: np.ndarray,
#     mask_flat: np.ndarray,
#     bead_mask_flat: np.ndarray | None = None,
# ) -> pl.DataFrame | pl.LazyFrame:
#     img_flat = img.ravel()

#     ## optional masking with beads
#     if bead_mask_flat is not None:
#         bead_mask_flat = ~bead_mask_flat
#         mask_flat = mask_flat[bead_mask_flat]
#         img_flat = img_flat[bead_mask_flat]

#     # intensity_means = np.bincount(mask_flat, weights=img_flat)[1:] / np.bincount(mask_flat)[1:]
#     counts = np.bincount(mask_flat)[1:]
#     sums = np.bincount(mask_flat, weights=img_flat)[1:]

#     # Avoid division by zero:
#     intensity_means = np.divide(sums, counts, out=np.zeros_like(sums), where=counts != 0)
#     return intensity_means


def image_mask_intensity_extraction_v2(
    img: np.ndarray,
    mask_flat: np.ndarray,
    bead_mask_flat: np.ndarray | None = None,
) -> np.ndarray:
    img_flat = img.ravel()

    # Optional bead masking
    if bead_mask_flat is not None:
        bead_mask_flat = ~bead_mask_flat
        mask_flat = mask_flat[bead_mask_flat]
        img_flat = img_flat[bead_mask_flat]

    # Remove zeros and find unique labels
    mask_nonzero = mask_flat > 0
    labels = mask_flat[mask_nonzero]
    pixels = img_flat[mask_nonzero]

    unique_labels, inv = np.unique(labels, return_inverse=True)
    # `inv` now contains remapped labels 0..n_unique-1

    # Compute sums and counts using remapped labels
    sums = np.bincount(inv, weights=pixels)
    counts = np.bincount(inv)

    # Safe divide
    means = np.divide(sums, counts, out=np.zeros_like(sums), where=counts != 0)

    return means


def vectorize_mask(mask: np.ndarray, nudge: bool = True) -> GeoDataFrame:
    if mask.max() == 0:
        return GeoDataFrame(geometry=[])

    regions = skimage.measure.regionprops(mask)

    geoms = []
    labels = []
    # Wrap the iteration in tqdm to show progress
    for region in tqdm(regions, desc='Vectorizing regions'):
        polys = region_props_to_polygons(region)
        geoms.extend(polys)
        # add the region label once per polygon
        labels.extend([region.label] * len(polys))

    gdf = GeoDataFrame({'label': labels}, geometry=geoms)

    if nudge:
        # GeoSeries.translate works elementwise
        gdf['geometry'] = gdf['geometry'].translate(xoff=-0.5, yoff=-0.5)

    return gdf


def region_props_to_polygons(region_props: RegionProperties) -> list[Polygon]:
    mask = np.pad(region_props.image, 1)
    contours = skimage.measure.find_contours(mask, 0.5)

    # shapes with <= 3 vertices, i.e. lines, can't be converted into a polygon
    polygons = [Polygon(contour[:, [1, 0]]) for contour in contours if contour.shape[0] >= 4]

    yoff, xoff, *_ = region_props.bbox
    return [shapely.affinity.translate(poly, xoff, yoff) for poly in polygons]


def rasterize_polygons(gdf: GeoDataFrame, target_shape: tuple) -> np.ndarray:
    height, width = target_shape
    transform = Affine.identity()

    # wrap the zip in tqdm; total=len(gdf) gives a proper progress bar
    wrapped = tqdm(zip(gdf.geometry, gdf['label']), total=len(gdf), desc='Rasterizing polygons')
    # feed that wrapped iterator into rasterize
    shapes = ((geom, int(lbl)) for geom, lbl in wrapped)

    label_array = rasterize(
        shapes=shapes,
        out_shape=(height, width),
        transform=transform,
        fill=0,  # background value
        dtype='int32',
    )

    return label_array



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class PointData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    POSITION_FIELD_NUMBER: builtins.int
    GENENAME_FIELD_NUMBER: builtins.int
    CELLID_FIELD_NUMBER: builtins.int
    geneName: builtins.str
    cellId: builtins.str
    @property
    def position(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        position: collections.abc.Iterable[builtins.float] | None = ...,
        geneName: builtins.str = ...,
        cellId: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal['cellId', b'cellId', 'geneName', b'geneName', 'position', b'position']
    ) -> None: ...

global___PointData = PointData

@typing.final
class TileData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    POINTSDATA_FIELD_NUMBER: builtins.int
    NUMBEROFPOINTS_FIELD_NUMBER: builtins.int
    numberOfPoints: builtins.int
    @property
    def pointsData(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___PointData]: ...
    def __init__(
        self,
        *,
        pointsData: collections.abc.Iterable[global___PointData] | None = ...,
        numberOfPoints: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal['numberOfPoints', b'numberOfPoints', 'pointsData', b'pointsData']
    ) -> None: ...

global___TileData = TileData

@typing.final
class ColumnData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COLUMNTILES_FIELD_NUMBER: builtins.int
    @property
    def columnTiles(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TileData]: ...
    def __init__(
        self,
        *,
        columnTiles: collections.abc.Iterable[global___TileData] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['columnTiles', b'columnTiles']) -> None: ...

global___ColumnData = ColumnData

@typing.final
class LevelData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEVELCOLUMNS_FIELD_NUMBER: builtins.int
    @property
    def levelColumns(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ColumnData]: ...
    def __init__(
        self,
        *,
        levelColumns: collections.abc.Iterable[global___ColumnData] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['levelColumns', b'levelColumns']) -> None: ...

global___LevelData = LevelData

@typing.final
class Transcripts(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEVEL_FIELD_NUMBER: builtins.int
    @property
    def level(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___LevelData]: ...
    def __init__(
        self,
        *,
        level: collections.abc.Iterable[global___LevelData] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal['level', b'level']) -> None: ...

global___Transcripts = Transcripts



import functools
import logging
import sys

from ..utils import LoggerWriter, setup_logger, validate_path


def workflow(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        wflow_name = func.__name__.removesuffix('_core')
        logger = kwargs.pop('logger', None)
        if logger is None:
            logger = setup_logger(logger_name=wflow_name, file_logger=False)
            logger.info('No logger provided, using default logger (stream only).')

        logger.info('-' * 10)
        logger.info(f'Initializing {wflow_name} workflow.')

        # Save the original stdout so we can restore it later
        original_stdout = sys.stdout
        sys.stdout = LoggerWriter(logger, logging.DEBUG)

        try:
            result = func(*args, logger=logger, **kwargs)
        except Exception as e:
            msg = f'Exception occurred during {wflow_name} workflow: \n\n{e}'
            logger.error(msg)
            raise RuntimeError(msg) from e
        else:
            logger.info(f'Completed {wflow_name} workflow.')
            return result
        finally:
            # Always restore stdout
            sys.stdout = original_stdout

    return wrapper


class OutSchema:
    def __init__(self, out_dir, subdirs=[]):
        self.out_dir = validate_path(out_dir, must_exist=False, is_dir_ok=True, is_file_ok=False, resolve_path=False)
        self.subdirs = subdirs

        for subdir in self.subdirs:
            p = self.out_dir / subdir
            p.mkdir(exist_ok=True, parents=True)
            setattr(self, subdir, p)



from __future__ import annotations

import json
import logging
import math
import multiprocessing as mp
import os
import shutil
import tarfile
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import polars as pl
from tqdm import tqdm

from ..g4x_viewer import TranscriptsSchema_pb2 as TranscriptsSchema
from .workflow import workflow

if TYPE_CHECKING:
    from ..models import G4Xoutput

mp.set_start_method('spawn', force=True)


@workflow
def create_tx_tarfile(
    g4x_obj: 'G4Xoutput',
    out_path: Path,
    *,
    tx_table: str | None = None,
    aggregation_level: str = 'gene',
    n_threads: int = 4,
    sampling_fraction: float = 0.2,
    logger: logging.Logger,
) -> None:
    logger.info('Generating viewer transcript file.')

    ## prelims
    IMAGE_RESOLUTION = g4x_obj.shape
    if g4x_obj.platform == 'g4x-2lane':
        MIN_TILE_SIZE = 1028
    else:
        MIN_TILE_SIZE = 512

    out_dir = out_path.parent / 'g4x_viewer_temp'
    os.makedirs(out_dir, exist_ok=True)

    ## get transcript table
    if aggregation_level == 'probe':
        tx_column = 'transcript'
    else:
        tx_column = 'gene_name'
    
    keep_cols = ['x_pixel_coordinate', 'y_pixel_coordinate', 'cell_id', tx_column]
    df = g4x_obj.load_transcript_table(lazy=True, columns=keep_cols, alt_file_path=tx_table)

    ## make colormap
    unique_genes = df.select(tx_column).collect().unique().to_series().to_list()
    num_genes = len(unique_genes)
    base_cmap = plt.get_cmap('hsv', num_genes)
    palette = {gene: [int(255 * c) for c in base_cmap(i / num_genes)[:3]] for i, gene in enumerate(unique_genes)}

    df = df.with_columns(
        (pl.col('x_pixel_coordinate') / MIN_TILE_SIZE).cast(pl.Int32).alias('tile_y_coord'),
        (pl.col('y_pixel_coordinate') / MIN_TILE_SIZE).cast(pl.Int32).alias('tile_x_coord'),
        pl.concat_list(['x_pixel_coordinate', 'y_pixel_coordinate']).alias('position'),
    )

    num_tiles_x, num_tiles_y = GetPyramidTilesConfigData(IMAGE_RESOLUTION, MIN_TILE_SIZE)
    NUMBER_OF_LEVELS = GetPyramidLevelsConfigData(IMAGE_RESOLUTION, MIN_TILE_SIZE)

    min_zoom_tiles_x = math.ceil(num_tiles_x / (pow(2, NUMBER_OF_LEVELS - 1)))
    min_zoom_tiles_y = math.ceil(num_tiles_y / (pow(2, NUMBER_OF_LEVELS - 1)))

    logger.info(f"""

        Final configurations for parsing:
            Image resolution: {IMAGE_RESOLUTION[0]} x {IMAGE_RESOLUTION[1]}
            Number of max zoom tiles: X = {num_tiles_x} | Y = {num_tiles_y}
            Number of min zoom tiles: X = {min_zoom_tiles_x} | Y = {min_zoom_tiles_y}
            Number of levels: {NUMBER_OF_LEVELS}
    """)

    save_configuration_file(out_dir, IMAGE_RESOLUTION, MIN_TILE_SIZE, NUMBER_OF_LEVELS, palette)
    logger.info('Parsing and classifying tiles...')

    pq_args = []
    for level_index in tqdm(reversed(range(NUMBER_OF_LEVELS + 1)), desc='Parsing and classifying tiles...'):
        # for level_index in reversed(range(NUMBER_OF_LEVELS + 1)):
        ## subsampling factor
        sampling_factor = sampling_fraction ** (NUMBER_OF_LEVELS - level_index)

        # factor for computing tile coordinates at this level
        scaling_factor = 2 ** (NUMBER_OF_LEVELS - level_index)
        current_tile_size = MIN_TILE_SIZE * scaling_factor
        x_num_of_tiles = math.ceil(IMAGE_RESOLUTION[0] / current_tile_size)
        y_num_of_tiles = math.ceil(IMAGE_RESOLUTION[1] / current_tile_size)

        # Ensure even numbers of tiles
        if x_num_of_tiles % 2 != 0:
            x_num_of_tiles += 1
        if y_num_of_tiles % 2 != 0:
            y_num_of_tiles += 1

        for tile_x_index in range(x_num_of_tiles):
            tileOutputDirPath = out_dir / f'{level_index}' / f'{tile_x_index}'
            os.makedirs(tileOutputDirPath, exist_ok=True)
            pq_args.append([tileOutputDirPath, y_num_of_tiles, scaling_factor])

            (
                df.filter(
                    ((pl.col('tile_x_coord') // scaling_factor) == tile_x_index),
                )
                .select(['position', tx_column, 'cell_id', 'tile_y_coord'])
                .collect()
                .sample(fraction=sampling_factor)
                .write_parquet(tileOutputDirPath / 'tmp.parquet')
            )

    logger.info('Processing transcripts')
    ctx = mp.get_context('spawn')
    with ctx.Pool(processes=n_threads) as pool:
        pool.starmap(_process_x, pq_args)

    logger.info('Tarring up.')
    if out_path.exists() or out_path.is_symlink():
        out_path.unlink()
    create_tar_from_directory(out_dir, out_path)
    shutil.rmtree(out_dir, ignore_errors=True)


def create_tar_from_directory(directory_path: str | Path, archive_name: str | Path) -> None:
    if not os.path.isdir(directory_path):
        raise ValueError(f'The directory {directory_path} does not exist.')
    with tarfile.open(archive_name, 'w') as tar:
        tar.add(directory_path, arcname=os.path.basename(directory_path))


def GetPyramidTilesConfigData(image_resolution: tuple[int, int], tile_size: int) -> tuple[int, int]:
    num_tiles_x = math.ceil(image_resolution[0] / tile_size)
    num_tiles_y = math.ceil(image_resolution[1] / tile_size)

    if not num_tiles_x % 2 == 0:
        num_tiles_x += 1

    if not num_tiles_y % 2 == 0:
        num_tiles_y += 1

    return num_tiles_x, num_tiles_y


def GetPyramidLevelsConfigData(image_resolution: tuple[int, int], tile_size: int, min_tiles_number: int = 16) -> int:
    min_num_levels = 0

    current_level = 1
    while min_num_levels == 0:
        level_tile_size = tile_size * pow(2, current_level)
        level_tiles_x = math.ceil(image_resolution[0] / level_tile_size)
        level_tiles_y = math.ceil(image_resolution[1] / level_tile_size)
        if not level_tiles_x % 2 == 0:
            level_tiles_x += 1
        if not level_tiles_y % 2 == 0:
            level_tiles_y += 1
        if level_tiles_x * level_tiles_y <= min_tiles_number:
            min_num_levels = current_level
            break
        current_level += 1

    return min_num_levels


def save_configuration_file(
    outputDirPath: str, image_resolution: tuple[int, int], min_tile_size: int, number_of_levels: int, palette: dict
) -> None:
    start_tile_size = min_tile_size * pow(2, number_of_levels)

    config_data = {
        'layer_height': image_resolution[0],
        'layer_width': image_resolution[1],
        'layers': number_of_levels,
        'tile_size': start_tile_size,
        'color_map': [{'gene_name': key, 'color': value} for key, value in palette.items()],
    }

    with open(f'{outputDirPath}/config.json', 'w') as json_file:
        json.dump(config_data, json_file, indent=2)


def _process_x(tileOutputDirPath: Path, y_num_of_tiles: int, scaling_factor: int) -> None:
    for tile_y_index in range(y_num_of_tiles):
        outputTileData = TranscriptsSchema.TileData()

        df_current = (
            pl.scan_parquet(tileOutputDirPath / 'tmp.parquet')
            .filter(((pl.col('tile_y_coord') // scaling_factor) == tile_y_index))
            .drop('tile_y_coord')
            .collect()
        )
        # Iterate over rows directly
        ## this can potentially be done lazily with this PR: https://github.com/pola-rs/polars/pull/23980
        for position, gene, cell_id in df_current.iter_rows():
            outputPointData = outputTileData.pointsData.add()
            _ = outputPointData.position.extend(position)
            outputPointData.geneName = gene
            outputPointData.cellId = str(cell_id)

        with open(f'{tileOutputDirPath}/{tile_y_index}.bin', 'wb') as file:
            _ = file.write(outputTileData.SerializeToString())



from .demux import demux_raw_features
from .edit_bin import edit_bin_file
from .init_bin import init_bin_file
from .segment import apply_segmentation
from .transcript_tar import create_tx_tarfile
from .viewer_dir import package_viewer_dir

__all__ = [
    'init_bin_file',
    'demux_raw_features',
    'apply_segmentation',
    'edit_bin_file',
    'create_tx_tarfile',
    'package_viewer_dir',
]



import importlib.resources as resources
import json
import logging
import os
import re
import shutil
from pathlib import Path

import polars as pl

from .. import utils
from ..models import G4Xoutput
from ..modules import create_tx_tarfile, edit_bin_file, init_bin_file
from ..modules.workflow import workflow
from . import validate

parquet_col_order = ['TXUID', 'sequence', 'confidence_score', 'x_pixel_coordinate', 'y_pixel_coordinate', 'z_level']


# register schemas
path = resources.files('g4x_helpers.schemas')
schemas = {}
for cont in path.iterdir():
    schemas[cont.name.removesuffix('.txt')] = cont


class MigrationError(Exception):
    pass


migration_targets = [
    ('g4x_viewer/{sample_id}.bin', 'g4x_viewer/{sample_id}_segmentation.bin', True),
    ('g4x_viewer/{sample_id}.tar', 'g4x_viewer/{sample_id}_transcripts.tar', True),
    ('g4x_viewer/{sample_id}.ome.tiff', 'g4x_viewer/{sample_id}_multiplex.ome.tiff', False),
    ('diagnostics/transcript_table.parquet', 'rna/raw_features.parquet', True),
    ('rna/transcript_table.parquet', 'rna/raw_features.parquet', True),
    ('rna/transcript_table.csv.gz', None, True),
    ('transcript_panel.csv', None, True),
    ('single_cell_data/feature_matrix.h5', None, True),
    ('single_cell_data/cell_metadata.csv.gz', None, True),
    ('g4x_viewer/{sample_id}_run_metadata.json', None, True),
]

backup_loc = 'g4x_helpers/migration_backup'


class MigrationTarget:
    def __init__(self, root: str, smp_id: str, source: str, target: str, backup: bool = True):
        self.root = Path(root)

        target = source if target is None else target
        source = source.format(sample_id=smp_id)
        target = target.format(sample_id=smp_id)

        self.is_rename = True if source != target else False
        self.rename_only = self.is_rename and backup is False

        self.source_file = self.root / source
        self.target_file = self.root / target

        self.backup = backup
        self.backup_dir = self.root / backup_loc
        self.backup_file = self.backup_dir / self.source_file.relative_to(self.root)

        self.source_file_short = self.source_file.relative_to(self.root)
        self.target_file_short = self.target_file.relative_to(self.root)
        self.backup_file_short = self.backup_file.relative_to(self.root)

    def check_source_exists(self) -> bool:
        return self.source_file.exists()

    def check_target_exists(self) -> bool:
        return self.target_file.exists()

    def check_backup_exists(self) -> bool:
        return self.backup_file.exists()

    def backup_source(self):
        if not self.backup:
            return
        if self.backup_file.exists():
            print(f'Backup file "{self.backup_file_short}" already exists. Will not overwrite.')
            return
        if self.source_file.exists():
            self.backup_file.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(self.source_file, self.backup_file)

    def restore_backup(self):
        if not self.backup:
            if self.check_target_exists() and self.is_rename:
                self.target_file.rename(self.source_file)
            return
        if not self.backup_file.exists():
            print(f'Can not restore file. Backup missing: "{self.backup_file_short}"')
            return
        self.target_file.unlink(missing_ok=False)
        self.backup_file.rename(self.source_file)

    def migrate(self):
        if not self.source_file.exists():
            print(f'Source file "{self.source_file_short}" does not exist. Cannot migrate.')
            return
        self.source_file.rename(self.target_file)

    def __repr__(self):
        return f'MigrationTarget: "{self.target_file_short}"'


@workflow
def migrate_g4x_data(
    data_dir: Path,
    sample_id: str,
    n_threads: int = utils.DEFAULT_THREADS,
    *,
    logger: logging.Logger,
):
    try:
        valid_schema = validate.validate_g4x_data(
            path=data_dir, schema_name='base_schema', formats={'sample_id': sample_id}, report=False
        )
        valid_files = validate.validate_file_schemas(data_dir)
        if not valid_schema or not valid_files:
            raise validate.ValidationError

        logger.info('All validations passed successfully. No migration needed.')
        return
    except Exception:
        logger.info('Migrating to latest G4X-data schema')

    backup_size = total_size_gb(migration_targets, sample_id=sample_id, base_path=data_dir)
    logger.info(f'Creating backup of {backup_size:.2f} GB before proceeding')

    ### STEP 1: Back up and migrate files to new locations
    logger.info('Updating file locations')
    mts = collect_targets(data_dir=data_dir, sample_id=sample_id)
    if any([mt.check_backup_exists() for mt in mts]):
        raise RuntimeError(
            'Non-empty backup directory detected. Will not proceed with migration to avoid overwriting existing backups. \n'
            'Please restore from backup via "g4x-helpers migrate --restore" or remove the backup directory if not needed.'
        )

    mt_sources = [mt for mt in mts if mt.check_source_exists()]

    for mt in mt_sources:
        mt.backup_source()
        logger.info(f'Migrating: {mt.source_file_short} -> {mt.target_file_short}')
        mt.migrate()

    logger.info('Validating results...')
    try:
        valid_schema = validate.validate_g4x_data(
            path=data_dir, schema_name='base_schema', formats={'sample_id': sample_id}, report=False
        )
    except validate.ValidationError:
        msg = (
            'Migration failed to produce correct G4X-data schema\n'
            'Your data was restored to its original state.\n'
            'Please run "g4x-helpers validate" to see which files are blocking validation.\n'
            'Contact care@singulargenomics.com for support'
        )
        restore(data_dir=data_dir, sample_id=sample_id, logger=logger)
        raise MigrationError(msg)

    logger.info('Successfully updated file locations! Checking file structures...')

    ### STEP 2: Validate file structures after migration
    logger.info('Checking for invalid NaN values in run_meta.json')
    fix_json_nan(data_dir / 'g4x_viewer' / f'{sample_id}_run_metadata.json')

    if not validate.validate_file_schemas(data_dir):
        logger.info('Some file structures need to be updated.')
        update_files = True
    else:
        logger.info('File structures are up to date! Migration complete.')
        return

    ### STEP 3: Update file structures as needed
    if update_files:
        parquet_lf, parquet_shema = validate.infer_parquet_schema(data_dir)
        logger.info(f'parquet schema is: {parquet_shema}')

        tx_panel_df, tx_panel_schema = validate.infer_tx_panel_schema(data_dir)
        logger.info(f'tx_panel schema is: {tx_panel_schema}')

        bin_file_schema = validate.infer_bin_schema(data_dir)
        logger.info(f'bin file schema is: {bin_file_schema}')

        adata, adata_schema = validate.infer_adata_schema(data_dir)
        logger.info(f'anndata schema is: {adata_schema}')

        if tx_panel_schema != 'valid' and tx_panel_schema != 'unknown':
            logger.info('Building new transcript_panel.csv from migrated transcript table')
            tx_panel = tx_panel_from_parquet(parquet_lf, tx_panel_df)
            tx_panel.write_csv(data_dir / 'transcript_panel.csv')

        if parquet_shema != 'valid' and parquet_shema != 'unknown':
            logger.info('Bringing parquet file in correct order')
            write_parquet_in_order(parquet_lf, data_dir)

        if adata_schema != 'valid' and adata_schema != 'unknown':
            logger.info('Bringing anndata file in correct order')
            write_adata_in_order(adata, data_dir)

        if bin_file_schema != 'valid' and bin_file_schema != 'unknown':
            smp = G4Xoutput(data_dir)
            init_bin_file(g4x_obj=smp, out_dir=data_dir, n_threads=n_threads, logger=logger)
            edit_bin_file(
                g4x_obj=smp, bin_file=data_dir / 'g4x_viewer' / f'{sample_id}_segmentation.bin', logger=logger
            )
            logger.info('Re-created bin file with updated schema.')
            create_tx_tarfile(
                g4x_obj=smp,
                out_path=data_dir / 'g4x_viewer' / f'{sample_id}_transcripts.tar',
                n_threads=n_threads,
                logger=logger,
            )

        logger.info('Re-validating file structures...')
        if not validate.validate_file_schemas(data_dir, verbose=True):
            msg = (
                'Migration failed to produce correct file structures.\n'
                'Your data was restored to its original state.\n'
                'Contact care@singulargenomics.com for support'
            )

            restore(data_dir=data_dir, sample_id=sample_id, logger=logger)
            raise MigrationError(msg)
        else:
            logger.info('File structures are up to date! Migration complete.')
            return True


@workflow
def restore_backup(
    data_dir: Path,
    sample_id: str,
    logger: logging.Logger,
):
    backup_dir = Path(data_dir) / backup_loc
    if backup_dir.exists():
        mts = collect_targets(data_dir=data_dir, sample_id=sample_id)
        mt_backups = [mt for mt in mts if mt.check_backup_exists() or mt.rename_only]

        if len(mt_backups) > 0:
            for mt in mt_backups:
                logger.info(f'Restoring: {mt.source_file_short}')
                mt.restore_backup()
        else:
            logger.info('No backed up files found. Nothing to restore.')
        shutil.rmtree(Path(data_dir) / backup_loc)
    else:
        logger.info('No backup found. Nothing to restore.')


def restore(
    data_dir: Path,
    sample_id: str,
    logger: logging.Logger,
):
    mts = collect_targets(data_dir=data_dir, sample_id=sample_id)
    mt_backups = [mt for mt in mts if mt.check_backup_exists() or mt.rename_only]

    if len(mt_backups) > 0:
        for mt in mt_backups:
            logger.info(f'Restoring: {mt.source_file_short}')
            mt.restore_backup()
    else:
        logger.info('No backed up files found. Nothing to restore.')
    shutil.rmtree(Path(data_dir) / backup_loc)


def collect_targets(data_dir: Path, sample_id: str):
    return [
        MigrationTarget(root=data_dir, smp_id=sample_id, source=mt[0], target=mt[1], backup=mt[2])
        for mt in migration_targets
    ]


def total_size_gb(files, sample_id=None, base_path='.'):
    total_bytes = 0

    for src, _, include in files:
        if not include:
            continue

        # Format path if sample_id is used
        if sample_id is not None:
            src = src.format(sample_id=sample_id)

        full_path = os.path.join(base_path, src)

        try:
            total_bytes += os.path.getsize(full_path)
        except FileNotFoundError:
            pass  # skip missing files

    return round(total_bytes / (1024**3), 2)


def fix_json_nan(path: Path) -> bool:
    with open(path) as f:
        txt = f.read()

    has_nan = bool(re.search(r'(?<!")\bNaN\b(?!")', txt))
    if has_nan:
        print('Found bare NaN. fixing...')

        # Replace bare NaN with "NaN"
        txt_fixed = txt.replace(' NaN', ' "NaN"').replace(':NaN', ':"NaN"')

        # Parse JSON safely now
        data = json.loads(txt_fixed)

        with open(path, 'w') as f:
            json.dump(data, f, indent=2)


def tx_panel_from_parquet(parquet_lf, tx_panel_old):
    tx_panel = (
        parquet_lf.filter(pl.col('probe_name') != 'UNDETERMINED')
        .unique('probe_name')
        .sort('probe_name')
        .select('probe_name', 'gene_name')
        .collect()
    )

    tx_panel = (
        tx_panel.unique('probe_name')
        .join(tx_panel_old, left_on='gene_name', right_on='target_condensed', how='right')
        .rename({'target_condensed': 'gene_name'})
        .sort('panel_type', 'gene_name')
        .fill_null('NOT_CONVERTED')
    )

    unconverted = tx_panel.filter(pl.col('probe_name') == 'NOT_CONVERTED')['gene_name']
    if len(unconverted) > 0:
        print(f'Warning: {len(unconverted)} gene_names failed to map to probe_names:')
        print(unconverted.to_list())
        print('They were listed in original transcript_panel.csv but not found in transcript data.')

    return tx_panel


def write_parquet_in_order(parquet_lf, sample_base):
    raw_features_path = sample_base / 'rna' / 'raw_features.parquet'
    tmp_file = raw_features_path.with_name('tmp.parquet')
    parquet_lf.select(parquet_col_order).sink_parquet(tmp_file)
    shutil.move(tmp_file, raw_features_path)


def write_adata_in_order(adata, sample_base: Path) -> None:
    adata_path = sample_base / 'single_cell_data' / 'feature_matrix.h5'
    cell_meta_path = sample_base / 'single_cell_data' / 'cell_metadata.csv.gz'

    cols = adata.obs.columns
    first_cols = ['nuclearstain_intensity_mean', 'cytoplasmicstain_intensity_mean']
    prot_cols = first_cols + [c for c in cols if '_intensity_mean' in c and c not in first_cols]
    cell_meta_cols = ['cell_id', 'cell_x', 'cell_y', 'nuclei_area_um', 'nuclei_expanded_area_um']
    last_cols = [c for c in cols if c not in cell_meta_cols and c not in prot_cols]

    new_cols = cell_meta_cols + prot_cols + last_cols

    adata.write_h5ad(adata_path)
    cell_metadata = adata.obs[new_cols].rename(columns={'cell_id': 'label'})
    utils.write_csv_gz(cell_metadata, cell_meta_path.with_suffix(''))



import json
import logging
import math
import shutil
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import polars as pl
from tqdm import tqdm

from .. import utils
from .workflow import OutSchema, workflow

if TYPE_CHECKING:
    from ..models import G4Xoutput

BASE_ORDER = 'CTGA'
LUT = np.zeros((256, 4), dtype=np.float32)
LUT[ord('C'), 0] = 1.0
LUT[ord('T'), 1] = 1.0
LUT[ord('G'), 2] = 1.0
LUT[ord('A'), 3] = 1.0


@workflow
def demux_raw_features(
    g4x_obj: 'G4Xoutput',
    manifest: Path,
    out_dir: Path,
    *,
    batch_size: int = 1_000_000,
    logger: logging.Logger,
) -> None:
    logger.info('Validating input paths.')
    out_tree = OutSchema(out_dir, subdirs=['g4x_viewer', 'rna', 'demux_batches'])
    manifest = utils.validate_path(manifest, must_exist=True, is_dir_ok=False, is_file_ok=True)
    tx_table_out = out_tree.rna / 'transcript_table.csv'

    ## update metadata and transcript panel file
    logger.info('Updating metadata and transcript panel file.')
    update_metadata_and_tx_file(g4x_obj, manifest, out_dir)

    logger.info('Parsing input transcript manifest.')
    manifest = utils.parse_input_manifest(file_path=manifest)

    logger.info('Running batched re-demuxing.')
    batched_demuxing(
        feature_table_path=g4x_obj.feature_table_path,
        manifest=manifest,
        batch_dir=out_tree.demux_batches,
        batch_size=batch_size,
    )

    logger.info('Compiling transcript table.')
    tx_table = pl.scan_parquet(list(out_tree.demux_batches.glob('*.parquet')))
    tx_table = tx_table.filter(pl.col('demuxed')).drop('demuxed')

    logger.info('Writing updated transcript table.')
    utils.write_csv_gz(tx_table, tx_table_out)

    logger.info('Cleaning up temporary files.')
    shutil.rmtree(out_tree.demux_batches)


def batched_demuxing(feature_table_path, manifest, batch_dir, batch_size):
    probe_dict = dict(zip(manifest['probe_name'].to_list(), manifest['gene_name'].to_list()))
    probe_dict['UNDETERMINED'] = 'UNDETERMINED'

    seq_reads = manifest['read'].unique().to_list()
    seq_reads = [int(x.split('_')[-1]) if isinstance(x, str) else x for x in seq_reads]

    num_features = pl.scan_parquet(feature_table_path).select(pl.len()).collect().item()
    num_expected_batches = math.ceil(num_features / batch_size)

    cols_to_select = ['TXUID', 'sequence', 'confidence_score', 'y_pixel_coordinate', 'x_pixel_coordinate', 'z_level']

    for i, feature_batch in tqdm(
        enumerate(stream_features(feature_table_path, batch_size, cols_to_select)),
        total=num_expected_batches,
        desc='Demuxing transcripts',
        position=0,
    ):
        feature_batch = feature_batch.with_columns(pl.col('TXUID').str.split('_').list.last().cast(int).alias('read'))
        redemuxed_feature_batch = []
        for seq_read in seq_reads:
            feature_batch_read = feature_batch.filter(pl.col('read') == seq_read)
            manifest_read = manifest.filter(pl.col('read') == seq_read)

            if len(feature_batch_read) == 0 or len(manifest_read) == 0:
                continue

            seqs = feature_batch_read['sequence'].to_list()
            codes = manifest_read['sequence'].to_list()
            codebook_target_ids = np.array(manifest_read['probe_name'].to_list())

            hammings = batched_dot_product_hamming_matrix(seqs, codes, batch_size=batch_size)
            feature_batch_read = demux(hammings, feature_batch_read, codebook_target_ids, probe_dict)
            feature_batch_read = feature_batch_read.drop(['sequence', 'read'])
            redemuxed_feature_batch.append(feature_batch_read)

        pl.concat(redemuxed_feature_batch).write_parquet(batch_dir / f'batch_{i}.parquet')


def stream_features(
    feature_table_path, batch_size: int, columns: str | list[str] | None = None
) -> Iterator[pl.DataFrame]:
    df = pl.scan_parquet(feature_table_path)
    if columns:
        df = df.select(columns)
    offset = 0
    while True:
        batch = df.slice(offset, batch_size).collect()
        if batch.is_empty():
            break
        yield batch
        offset += batch_size


def demux(
    hammings: np.ndarray,
    reads: pl.DataFrame,
    codebook_target_ids: np.ndarray,
    probe_dict: dict,
    max_ham_dist: int = 2,
    min_delta: int = 2,
) -> pl.DataFrame:
    demuxed = np.zeros(hammings.shape[0], dtype=bool)

    for i in range(max_ham_dist + 1):
        hits = hammings == i
        close_hits = hammings <= (i + min_delta)
        uniquely_hit = hits.sum(axis=1) == 1
        close_hit = close_hits.sum(axis=1) > 1
        pass_filter = uniquely_hit & ~close_hit
        demuxed[pass_filter] = 1

        # logger.info(f"""
        # ... ... {fq.stem}
        # hamming == {i}, min_delta == {min_delta}
        # unique hits = {sum(uniquely_hit)}
        # total cumulative hits within min_delta = {sum(close_hit)}
        # total demuxed (unique hits without another hit within min_delta) = {sum(pass_filter)}
        # """)

    # --- Get best hits ---
    hit_ids = hammings.argmin(axis=1)
    hit_targets = codebook_target_ids[hit_ids]

    transcripts = np.where(demuxed, hit_targets, 'UNDETERMINED')
    transcript_condensed = [probe_dict.get(t, 'UNDETERMINED') for t in transcripts]

    reads = reads.with_columns(
        [
            pl.Series('probe_name', transcripts),
            pl.Series('gene_name', transcript_condensed),
            pl.Series('demuxed', demuxed),
        ]
    )

    return reads


def batched_dot_product_hamming_matrix(
    reads: list[str],
    codebook: list[str],
    batch_size: int,
) -> np.ndarray:
    """
    Compute full Hamming distance matrix (N_reads, N_codebook)
    using batched dot-product with one-hot encoding.
    """
    seq_len = len(codebook[0])
    assert all(len(seq) == seq_len for seq in codebook), 'All codebook entries must be same length'

    # One-hot encode the codebook once
    codebook_oh = one_hot_encode_str_array(codebook, seq_len)
    M = len(codebook)

    # Prepare final result
    N = len(reads)
    hamming_matrix = np.empty((N, M), dtype=np.uint8)

    num_expected_batches = math.ceil(N / batch_size)
    for i in tqdm(range(0, N, batch_size), total=num_expected_batches, desc='Demuxing batch', position=1, leave=False):
        batch_reads = reads[i : i + batch_size]
        batch_oh = one_hot_encode_str_array(batch_reads, seq_len)
        matches = batch_oh @ codebook_oh.T
        hamming = seq_len - matches
        hamming_matrix[i : i + len(batch_reads)] = hamming

    return hamming_matrix


def one_hot_encode_str_array(seqs: list[str], seq_len: int) -> np.ndarray:
    """
    Fast one-hot encoding using LUT.
    Returns: (N, seq_len * 4) float32 array
    """
    N = len(seqs)
    # Flatten all sequences into a byte array and reshape to (N, seq_len)
    arr = np.frombuffer(''.join(seqs).encode('ascii'), dtype=np.uint8).reshape(N, seq_len)
    # Apply LUT: arr → (N, seq_len, 4), then flatten
    return LUT[arr].reshape(N, seq_len * 4)


def update_metadata_and_tx_file(g4x_obj: 'G4Xoutput', manifest, out_dir):
    if not manifest == out_dir / 'transcript_panel.csv':
        shutil.copy(manifest, out_dir / 'transcript_panel.csv')

    panel_name = manifest.name
    timestamp = f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}'
    ## add info to run_meta.json
    with open(g4x_obj.data_dir / 'run_meta.json', 'r') as f:
        meta = json.load(f)

    meta['transcript_panel'] = panel_name
    meta['redemuxed_timestamp'] = timestamp

    with open(out_dir / 'run_meta.json', 'w') as f:
        json.dump(meta, f, indent=2)

    ## add info to run_meta.json in g4x_viewer
    with open(g4x_obj.data_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_run_metadata.json', 'r') as f:
        meta = json.load(f)

    meta['run_metadata']['transcript_panel'] = panel_name
    meta['run_metadata']['redemuxed_timestamp'] = timestamp

    with open(out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_run_metadata.json', 'w') as f:
        json.dump(meta, f, indent=2)


def concatenate_and_cleanup(batch_dir, out_dir):
    final_tx_table_path = out_dir / 'rna' / 'transcript_table.csv'
    # final_tx_pq_path = out_dir / 'rna' / 'transcript_table.parquet'

    utils.delete_existing(final_tx_table_path)
    utils.delete_existing(final_tx_table_path.with_suffix('.csv.gz'))
    # utils.delete_existing(final_tx_pq_path)

    tx_table = pl.scan_parquet(list(batch_dir.glob('*.parquet')))
    # tx_table.sink_parquet(final_tx_pq_path)

    tx_table.filter(pl.col('demuxed')).sink_csv(final_tx_table_path)

    _ = utils.gzip_file(final_tx_table_path)
    shutil.rmtree(batch_dir)



CLI_HELP = 'Helper models and post-processing tools for G4X-data\n\ndocs.singulargenomics.com'

RESEG_HELP = (
    'Reprocess G4X-data with a new segmentation\n\n'
    'Takes new cell-labels from a custom segmentation output and re-assigns transcripts and protein signals to those cells. '
    'The operation recreates most single-cell outputs and initializes a new G4X-viewer "segmentation.bin" file. '
    'Does not regenerate umaps/clustering or metrics.'
)

REDMX_HELP = (
    'Reprocess G4X-data with a new transcript manifest\n\n'
    'Generates a new "transcript_table.csv" by demultiplexing the raw feature data against a provided list of probe sequences '
    'and mapping each feature to its corresponding gene/target name. It then proceeds to regenerate single-cell outputs and initializes '
    'a new G4X-viewer "segmentation.bin" and "transcripts.tar" file.'
)


UDBIN_HELP = (
    'Update existing G4X-viewer segmentation file with new metadata\n\n'
    'Populates cluster or embedding information in an existing G4X-viewer segmentation.bin file. '
    'Only updates fields that are provided leaving other fields unchanged. Cell-IDs in the metadata table must match those in the bin file. '
    'Any cell-IDs in the metadata that are not found in the bin file will be assigned a default value.'
)

NWBIN_HELP = (
    'Initialize new segmentation file for G4X-viewer\n\n '
    'Converts cell-labels from "segmentation_mask.npz" into a "segmentation.bin" file compatible with the latest version of G4X-viewer. '
    'Cluster labels and umap embeddings are not populated in this operation. Please use "g4x-helpers update_bin" to add desired metadata.\n\n'
)

TARVW_HELP = (
    'Package G4X-viewer folder for distribution\n\n'
    'Creates a .tar archive of the "g4x_viewer" directory for easy upload and sharing.\n\n'
    'Archive file name: {SAMPLE_ID}_viewer.tar\n\n'
)

MIGRT_HELP = (
    'Migrate legacy data to comply with latest schemas for G4X-viewer & helpers\n\n'
    'Updates file names, headers and locations and creates new G4X-viewer files if necessary.\n\n'
    'A backup of the originals is created before migrating and operation will not proceed if an existing backup is present.\n\n'
)

VLDTE_HELP = 'Validate G4X-data to ensure correct file and folder structure\n\n'
